@@ -48,6 +48,16 @@
 #endif
 #endif
 
+#ifndef __SWORD_TYPE
+#if __WORDSIZE == 32
+# define __SWORD_TYPE	int
+#elif __WORDSIZE == 64
+# define __SWORD_TYPE	long int
+#else
+#error
+#endif
+#endif
+
 const char *program;		/* Initialized with argv[0] */
 const char *version = VERSION_STRING;	/* Program version */
 const char *libdir = AUTOFS_LIB_DIR;	/* Location of library modules */
@@ -59,9 +69,15 @@ unsigned int mp_mode = 0755;
 unsigned int nfs_mount_uses_string_options = 0;
 static struct nfs_mount_vers vers, check = {1, 1, 1};
 
-/* autofs fifo name prefix */
-#define FIFO_NAME_PREFIX "autofs.fifo"
-const char *fifodir = AUTOFS_FIFO_DIR "/" FIFO_NAME_PREFIX;
+#define FIFO_BUF_SIZE		25
+static int cmd_pipe_fifo = -1;
+
+/* autofs cmd fifo name */
+#define FIFO_NAME "autofs.cmd.fifo"
+const char *cmd_pipe_name = AUTOFS_FIFO_DIR "/" FIFO_NAME;
+
+int start_cmd_pipe_handler(void);
+void finish_cmd_pipe_handler(void);
 
 const char *global_options;		/* Global option, from command line */
 
@@ -69,12 +85,13 @@ static char *pid_file = NULL;		/* File in which to keep pid */
 unsigned int global_selection_options;
 
 long global_negative_timeout = -1;
+long global_positive_timeout = -1;
 int do_force_unlink = 0;		/* Forceably unlink mount tree at startup */
 
 static int start_pipefd[2] = {-1, -1};
 static int st_stat = 1;
 static int *pst_stat = &st_stat;
-static pthread_t state_mach_thid;
+static pthread_t signal_handler_thid;
 
 static sigset_t block_sigs;
 
@@ -94,8 +111,6 @@ struct startup_cond suc = {
 pthread_key_t key_thread_stdenv_vars;
 pthread_key_t key_thread_attempt_id = (pthread_key_t) 0L;
 
-#define MAX_OPEN_FILES		10240
-
 int aquire_flag_file(void);
 void release_flag_file(void);
 
@@ -558,7 +573,7 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 		struct mapent *tmp;
 		int ret;
 
-		ret = tree_mapent_umount_offsets(me, 1);
+		ret = tree_mapent_umount_offsets(me);
 		if (!ret) {
 			warn(ap->logopt,
 			     "some offset mounts still present under %s", path);
@@ -609,8 +624,7 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 		struct mnt_list *mnt;
 
 		debug(ap->logopt, "unmounting dir = %s", path);
-		if (umount_ent(ap, path) &&
-		    is_mounted(path, MNTS_REAL)) {
+		if (umount_ent(ap, path)) {
 			warn(ap->logopt, "could not umount dir %s", path);
 			left++;
 			goto done;
@@ -705,13 +719,6 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 
 	left = 0;
 
-	/*
-	 * If we are a submount we need to umount any offsets our
-	 * parent may have mounted over top of us.
-	 */
-	if (ap->submount)
-		left += umount_subtree_mounts(ap->parent, path, 1);
-
 	left += umount_subtree_mounts(ap, path, is_autofs_fs);
 
 	/* Delete detritus like unwanted mountpoints and symlinks */
@@ -735,23 +742,16 @@ static void umount_all(struct autofs_point *ap)
 		     left, ap->path);
 }
 
-static int umount_autofs(struct autofs_point *ap, const char *root)
+static int umount_autofs(struct autofs_point *ap)
 {
 	int ret = 0;
 
 	if (ap->state == ST_INIT)
 		return -1;
 
-	/*
-	 * Since lookup.c is lazy about closing lookup modules
-	 * to prevent unneeded opens, we need to clean them up
-	 * before umount.
-	 */
-	lookup_close_lookup(ap);
-
 	if (ap->type == LKP_INDIRECT) {
 		umount_all(ap);
-		ret = umount_autofs_indirect(ap, root);
+		ret = umount_autofs_indirect(ap);
 	} else
 		ret = umount_autofs_direct(ap);
 
@@ -803,333 +803,47 @@ static int fullread(int fd, void *ptr, size_t len)
 	return len;
 }
 
-static char *automount_path_to_fifo(unsigned logopt, const char *path)
+static void dummy(int sig)
 {
-	char *fifo_name, *p;
-	int  name_len = strlen(path) + strlen(fifodir) + 1;
-	int ret;
-
-	fifo_name = malloc(name_len);
-	if (!fifo_name)
-		return NULL;
-	ret = snprintf(fifo_name, name_len, "%s%s", fifodir, path);
-	if (ret >= name_len) {
-		info(logopt,
-		     "fifo path for \"%s\" truncated to \"%s\".  This may "
-		     "lead to --set-log-priority commands being sent to the "
-		     "wrong automount daemon.", path, fifo_name);
-	}
-
-	/*
-	 *  An automount path can be made up of subdirectories.  So, to
-	 *  create the fifo name, we will just replace instances of '/' with
-	 *  '-'. 
-	 */
-	p = fifo_name + strlen(fifodir);
-	while (*p != '\0') {
-		if (*p == '/')
-			*p = '-';
-		p++;
-	}
-
-	debug(logopt, "fifo name %s",fifo_name);
-
-	return fifo_name;
-}
-
-static int create_logpri_fifo(struct autofs_point *ap)
-{
-	int ret = -1;
-	int fd;
-	char *fifo_name;
-	char buf[MAX_ERR_BUF];
-
-	fifo_name = automount_path_to_fifo(ap->logopt, ap->path);
-	if (!fifo_name) {
-		crit(ap->logopt, "Failed to allocate memory!");
-		goto out_free; /* free(NULL) is okay */
-	}
-
-	ret = unlink(fifo_name);
-	if (ret != 0 && errno != ENOENT) {
-		crit(ap->logopt,
-		     "Failed to unlink FIFO. Is the automount daemon "
-		     "already running?");
-		goto out_free;
-	}
-
-	ret = mkfifo(fifo_name, S_IRUSR|S_IWUSR);
-	if (ret != 0) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		crit(ap->logopt,
-		     "mkfifo for %s failed: %s", fifo_name, estr);
-		goto out_free;
-	}
-
-	fd = open_fd(fifo_name, O_RDWR|O_NONBLOCK);
-	if (fd < 0) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		crit(ap->logopt,
-		     "Failed to open %s: %s", fifo_name, estr);
-		unlink(fifo_name);
-		ret = -1;
-		goto out_free;
-	}
-
-	ap->logpri_fifo = fd;
-
-out_free:
-	free(fifo_name);
-	return ret;
-}
-
-int destroy_logpri_fifo(struct autofs_point *ap)
-{
-	int ret = -1;
-	int fd = ap->logpri_fifo;
-	char *fifo_name;
-	char buf[MAX_ERR_BUF];
-
-	if (fd == -1)
-		return 0;
-
-	fifo_name = automount_path_to_fifo(ap->logopt, ap->path);
-	if (!fifo_name) {
-		crit(ap->logopt, "Failed to allocate memory!");
-		goto out_free; /* free(NULL) is okay */
-	}
-
-	ap->logpri_fifo = -1;
-
-	ret = close(fd);
-	if (ret != 0) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		warn(ap->logopt,
-		     "close for fifo %s: %s", fifo_name, estr);
-	}
-
-	ret = unlink(fifo_name);
-	if (ret != 0) {
-		warn(ap->logopt,
-		     "Failed to unlink FIFO. Was the fifo created OK?");
-	}
-
-out_free:
-	free(fifo_name);
-	return ret;
-}
-
-static void cleanup_stale_logpri_fifo_pipes(void)
-{
-	size_t prefix_len = strlen(FIFO_NAME_PREFIX);
-	char *dir = AUTOFS_FIFO_DIR;
-	size_t dir_len = strlen(dir);
-	struct dirent *dent;
-	DIR *dfd;
-	int ret;
-
-	dfd = opendir(dir);
-	if (!dfd) {
-		warn(LOGOPT_ANY, "failed to open fifo dir %s", dir);
-		return;
-	}
-
-	while ((dent = readdir(dfd))) {
-		char fifo_path[PATH_MAX];
-
-		if (!(dent->d_type & DT_FIFO))
-			continue;
-		if (strncmp(FIFO_NAME_PREFIX, dent->d_name, prefix_len))
-			continue;
-		if ((dir_len + 1 + strlen(dent->d_name)) >= PATH_MAX) {
-			warn(LOGOPT_ANY, "fifo path too long for buffer");
-			continue;
-		}
-
-		strcpy(fifo_path, dir);
-		strcat(fifo_path, "/");
-		strcat(fifo_path, dent->d_name);
-
-		ret = unlink(fifo_path);
-		if (ret == -1) {
-			char buf[MAX_ERR_BUF];
-			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			warn(LOGOPT_ANY, "unlink of fifo failed: %s", estr);
-		}
-	}
-
-	closedir(dfd);
-}
-
-static void handle_fifo_message(struct autofs_point *ap, int fd)
-{
-	int ret;
-	char buffer[PIPE_BUF];
-	char *end;
-	long pri;
-	char buf[MAX_ERR_BUF];
-
-	memset(buffer, 0, sizeof(buffer));
-	ret = read(fd, &buffer, sizeof(buffer));
-	if (ret < 0) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		warn(ap->logopt, "read on fifo returned error: %s", estr);
-		return;
-	}
-
-	if (ret != 2) {
-		debug(ap->logopt, "expected 2 bytes, received %d.", ret);
-		return;
-	}
-
-	errno = 0;
-	pri = strtol(buffer, &end, 10);
-	if ((pri == LONG_MIN || pri == LONG_MAX) && errno == ERANGE) {
-		debug(ap->logopt, "strtol reported an %s.  Failed to set "
-		      "log priority.", pri == LONG_MIN ? "underflow" : "overflow");
-		return;
-	}
-	if ((pri == 0 && errno == EINVAL) || end == buffer) {
-		debug(ap->logopt, "priority is expected to be an integer "
-		      "in the range 0-7 inclusive.");
-		return;
-	}
-
-	if (pri > LOG_DEBUG || pri < LOG_EMERG) {
-		debug(ap->logopt, "invalid log priority (%ld) received "
-		      "on fifo", pri);
-		return;
-	}
-
-	/*
-	 * OK, the message passed all of the sanity checks.  The
-	 * automounter actually only supports three log priorities.
-	 * Everything is logged at log level debug, deamon messages
-	 * and everything except debug messages are logged with the
-	 * verbose setting and only error and critical messages are
-	 * logged when debugging isn't enabled.
-	 */
-	if (pri >= LOG_WARNING) {
-		if (pri == LOG_DEBUG) {
-			set_log_debug_ap(ap);
-			info(ap->logopt, "Debug logging set for %s", ap->path);
-		} else {
-			set_log_verbose_ap(ap);
-			info(ap->logopt, "Verbose logging set for %s", ap->path);
-		}
-	} else {
-		if (ap->logopt & LOGOPT_ANY)
-			info(ap->logopt, "Basic logging set for %s", ap->path);
-		set_log_norm_ap(ap);
-	}
-}
-
-static int set_log_priority(const char *path, int priority)
-{
-	int fd;
-	char *fifo_name;
-	char buf[2];
-
-	if (priority > LOG_DEBUG || priority < LOG_EMERG) {
-		fprintf(stderr, "Log priority %d is invalid.\n", priority);
-		fprintf(stderr, "Please specify a number in the range 0-7.\n");
-		return -1;
-	}
-
-	/*
-	 * This is an ascii based protocol, so we want the string
-	 * representation of the integer log priority.
-	 */
-	snprintf(buf, sizeof(buf), "%d", priority);
-
-	fifo_name = automount_path_to_fifo(LOGOPT_NONE, path);
-	if (!fifo_name) {
-		fprintf(stderr, "%s: Failed to allocate memory!\n",
-			__FUNCTION__);
-		return -1;
-	}
-
-	/*
-	 * Specify O_NONBLOCK so that the open will fail if there is no
-	 * daemon reading from the other side of the FIFO.
-	 */
-	fd = open_fd(fifo_name, O_WRONLY|O_NONBLOCK);
-	if (fd < 0) {
-		fprintf(stderr, "%s: open of %s failed with %s\n",
-			__FUNCTION__, fifo_name, strerror(errno));
-		fprintf(stderr, "%s: perhaps the fifo wasn't setup,"
-			" please check your log for more information\n", __FUNCTION__);
-		free(fifo_name);
-		return -1;
-	}
-
-	if (write(fd, buf, sizeof(buf)) != sizeof(buf)) {
-		fprintf(stderr, "Failed to change logging priority.  ");
-		fprintf(stderr, "write to fifo failed: %s.\n",
-			strerror(errno));
-		close(fd);
-		free(fifo_name);
-		return -1;
-	}
-	close(fd);
-	free(fifo_name);
-	fprintf(stdout, "Successfully set log priority for %s.\n", path);
-
-	return 0;
 }
 
 static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 {
-	struct pollfd fds[3];
-	int pollfds = 3;
+	struct sigaction sa;
+	sigset_t signalset;
+	struct pollfd fds[1];
+	int pollfds = 1;
 	char buf[MAX_ERR_BUF];
 	size_t read;
 	char *estr;
 
 	fds[0].fd = ap->pipefd;
 	fds[0].events = POLLIN;
-	fds[1].fd = ap->state_pipe[0];
-	fds[1].events = POLLIN;
-	fds[2].fd = ap->logpri_fifo;
-	fds[2].events = POLLIN;
-	if (fds[2].fd  == -1)
-		pollfds--;
+
+	sa.sa_handler = dummy;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = 0;
+	if (sigaction(SIGCONT, &sa, NULL) == -1)
+		error(LOGOPT_ANY, "failed to set signal handler %d", errno);
+
+	sigfillset(&signalset);
+	sigdelset(&signalset, SIGCONT);
 
 	for (;;) {
-		if (poll(fds, pollfds, -1) == -1) {
-			if (errno == EINTR)
-				continue;
-			estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			logerr("poll failed: %s", estr);
-			return -1;
-		}
-
-		if (fds[1].revents & POLLIN) {
-			enum states next_state;
-			size_t read_size = sizeof(next_state);
-			int state_pipe;
-
-			next_state = ST_INVAL;
-
-			st_mutex_lock();
-
-			state_pipe = ap->state_pipe[0];
-
-			read = fullread(state_pipe, &next_state, read_size);
-			if (read) {
-				estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				error(ap->logopt,
-				      "read error on state pipe, "
-				      "read %lu, error %s",
-				      read, estr);
+		errno = 0;
+		if (ppoll(fds, pollfds, NULL, &signalset) == -1) {
+			if (errno == EINTR) {
+				st_mutex_lock();
+				if (ap->state == ST_SHUTDOWN) {
+					st_mutex_unlock();
+					return -1;
+				}
 				st_mutex_unlock();
 				continue;
 			}
-
-			st_mutex_unlock();
-
-			if (next_state == ST_SHUTDOWN)
-				return -1;
+			estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			logerr("poll failed: %s", estr);
+			return -1;
 		}
 
 		if (fds[0].revents & POLLIN) {
@@ -1143,25 +857,23 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 			}
 			return read;
 		}
-
-		if (fds[2].fd != -1 && fds[2].revents & POLLIN) {
-			debug(ap->logopt, "message pending on control fifo.");
-			handle_fifo_message(ap, fds[2].fd);
-		}
 	}
 }
 
 int do_expire(struct autofs_point *ap, const char *name, int namelen)
 {
 	char buf[PATH_MAX];
+	const char *parent;
 	int len, ret;
 
 	if (*name != '/') {
 		len = ncat_path(buf, sizeof(buf), ap->path, name, namelen);
+		parent = ap->path;
 	} else {
 		len = snprintf(buf, PATH_MAX, "%s", name);
 		if (len >= PATH_MAX)
 			len = 0;
+		parent = name;
 	}
 
 	if (!len) {
@@ -1169,13 +881,13 @@ int do_expire(struct autofs_point *ap, const char *name, int namelen)
 		return 1;
 	}
 
-	info(ap->logopt, "expiring path %s", buf);
+	info(ap->logopt, "expiring path %s on %s", buf, parent);
 
 	pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 	master_source_readlock(ap->entry);
 	ret = umount_multi(ap, buf, 1);
 	if (ret == 0)
-		info(ap->logopt, "expired %s", buf);
+		info(ap->logopt, "umounting %s succeeded", buf);
 	else
 		warn(ap->logopt, "couldn't complete expire of %s", buf);
 	pthread_cleanup_pop(1);
@@ -1206,26 +918,12 @@ static int autofs_init_ap(struct autofs_point *ap)
 	ap->pipefd = pipefd[0];
 	ap->kpipefd = pipefd[1];
 
-	/* Pipe state changes from signal handler to main loop */
-	if (open_pipe(ap->state_pipe) < 0) {
-		crit(ap->logopt,
-		     "failed create state pipe for autofs path %s", ap->path);
-		close(ap->pipefd);
-		close(ap->kpipefd);	/* Close kernel pipe end */
-		return -1;
-	}
-
-	if (create_logpri_fifo(ap) < 0) {
-		logmsg("could not create FIFO for path %s\n", ap->path);
-		logmsg("dynamic log level changes not available for %s", ap->path);
-	}
-
 	return 0;
 }
 
-static int mount_autofs(struct autofs_point *ap, const char *root)
+static int mount_autofs(struct autofs_point *ap)
 {
-	int status = 0;
+	int status;
 
 	/* No need to create comms fds and command fifo if
 	 * unlinking mounts and exiting.
@@ -1238,16 +936,11 @@ static int mount_autofs(struct autofs_point *ap, const char *root)
 	if (ap->type == LKP_DIRECT)
 		status = mount_autofs_direct(ap);
 	else
-		status = mount_autofs_indirect(ap, root);
-
-	if (status < 0) {
-		destroy_logpri_fifo(ap);
-		return -1;
-	}
+		status = mount_autofs_indirect(ap);
 
 	st_add_task(ap, ST_READY);
 
-	return 0;
+	return status;
 }
 
 static int handle_packet(struct autofs_point *ap)
@@ -1407,6 +1100,21 @@ static unsigned long getnumopt(char *str, char option)
 	return val;
 }
 
+static long getsnumopt(char *str, char option)
+{
+	long val;
+	char *end;
+
+	val = strtol(str, &end, 0);
+	if (!*str || *end) {
+		fprintf(stderr,
+			"%s: option -%c requires a numeric argument, got %s\n",
+			program, option, str);
+		exit(1);
+	}
+	return val;
+}
+
 static void do_master_cleanup_unlock(void *arg)
 {
 	int status;
@@ -1595,7 +1303,7 @@ static int do_hup_signal(struct master *master)
 }
 
 /* Deal with all the signal-driven events in the state machine */
-static void *statemachine(void *arg)
+static void *signal_handler(void *arg)
 {
 	sigset_t signalset;
 	int sig;
@@ -1642,6 +1350,337 @@ static void *statemachine(void *arg)
 	}
 }
 
+static pthread_mutex_t cmd_pipe_mutex = PTHREAD_MUTEX_INITIALIZER;
+static unsigned int done = 0;
+static pthread_t cmd_pipe_thid;
+
+void cmd_pipe_mutex_lock(void)
+{
+	int status = pthread_mutex_lock(&cmd_pipe_mutex);
+	if (status)
+		fatal(status);
+}
+
+void cmd_pipe_mutex_unlock(void)
+{
+	int status = pthread_mutex_unlock(&cmd_pipe_mutex);
+	if (status)
+		fatal(status);
+}
+
+static int create_cmd_pipe_fifo(void)
+{
+	char buf[MAX_ERR_BUF];
+	int ret = -1;
+	int fd;
+
+	if (cmd_pipe_fifo != -1)
+		return 0;
+
+	ret = unlink(cmd_pipe_name);
+	if (ret != 0 && errno != ENOENT) {
+		fprintf(stderr,
+			"%s: failed to unlink command pipe. Is the "
+			"automount daemon already running?", program);
+		return ret;
+	}
+
+	ret = mkfifo(cmd_pipe_name, S_IRUSR|S_IWUSR);
+	if (ret != 0 && errno != EEXIST) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		fprintf(stderr, "%s: mkfifo for %s failed: %s",
+			program, cmd_pipe_name, estr);
+		return ret;
+	}
+
+	fd = open_fd(cmd_pipe_name, O_RDWR|O_NONBLOCK);
+	if (fd < 0) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		unlink(cmd_pipe_name);
+		fprintf(stderr, "%s: failed to open command pipe %s: %s",
+			program, cmd_pipe_name, estr);
+		return -1;
+	}
+
+	cmd_pipe_fifo = fd;
+
+	return 0;
+}
+
+static int destroy_cmd_pipe_fifo(void)
+{
+	char buf[MAX_ERR_BUF];
+	int ret = -1;
+
+	if (cmd_pipe_fifo == -1)
+		return 0;
+
+	ret = close(cmd_pipe_fifo);
+	if (ret != 0) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		warn(LOGOPT_ANY,
+		     "close for command pipe %s: %s", cmd_pipe_name, estr);
+	}
+
+	cmd_pipe_fifo = -1;
+
+	ret = unlink(cmd_pipe_name);
+	if (ret != 0) {
+		warn(LOGOPT_ANY,
+		     "failed to unlink FIFO. Was the fifo created OK?");
+	}
+
+	return 0;
+}
+
+static void handle_cmd_pipe_fifo_message(int fd)
+{
+	struct autofs_point *ap;
+	char buffer[PIPE_BUF];
+	char *end, *sep;
+	char buf[MAX_ERR_BUF];
+	dev_t devid;
+	int ret;
+	long pri;
+
+	ret = read(fd, &buffer, sizeof(buffer));
+	if (ret < 0) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		warn(LOGOPT_ANY,
+		     "read on command pipe returned error: %s", estr);
+		return;
+	}
+	if (ret >= sizeof(buffer)) {
+		error(LOGOPT_ANY,
+		      "read overrun on command pipe message");
+		return;
+	}
+	buffer[ret] = 0;
+
+	sep = strrchr(buffer, ' ');
+	if (!sep) {
+		error(LOGOPT_ANY,
+		      "incorrect command pipe message format %s.", buffer);
+		return;
+	}
+	sep++;
+
+	errno = 0;
+	devid = strtol(buffer, &end, 10);
+	if ((devid == LONG_MIN || devid == LONG_MAX) && errno == ERANGE) {
+		debug(LOGOPT_ANY, "strtol reported a range error.");
+		error(LOGOPT_ANY, "invalid command pipe message format %s.", buffer);
+		return;
+	}
+
+	if ((devid == 0 && errno == EINVAL) || end == buffer) {
+		debug(LOGOPT_ANY, "devid id is expected to be a integer.");
+		return;
+	}
+
+	ap = master_find_mapent_by_devid(master_list, devid);
+	if (!ap) {
+		error(LOGOPT_ANY, "can't locate autofs_point for device id %ld.", devid);
+		return;
+	}
+
+	errno = 0;
+	pri = strtol(sep, &end, 10);
+	if ((pri == LONG_MIN || pri == LONG_MAX) && errno == ERANGE) {
+		error(ap->logopt, "failed to set log priority.");
+		error(ap->logopt, "strtol reported an %s.",
+		      pri == LONG_MIN ? "underflow" : "overflow");
+		return;
+	}
+
+	if ((pri == 0 && errno == EINVAL) || end == sep) {
+		debug(ap->logopt, "priority is expected to be an integer "
+		      "in the range 0-7 inclusive.");
+		return;
+	}
+
+	if (pri > LOG_DEBUG || pri < LOG_EMERG) {
+		debug(ap->logopt,
+		      "invalid log priority (%ld) received on fifo", pri);
+		return;
+	}
+
+	/*
+	 * OK, the message passed all of the sanity checks.  The
+	 * automounter actually only supports three log priorities.
+	 * Everything is logged at log level debug, deamon messages
+	 * and everything except debug messages are logged with the
+	 * verbose setting and only error and critical messages are
+	 * logged when debugging isn't enabled.
+	 */
+	if (pri >= LOG_WARNING) {
+		if (pri == LOG_DEBUG) {
+			set_log_debug_ap(ap);
+			info(ap->logopt, "debug logging set for %s", ap->path);
+		} else {
+			set_log_verbose_ap(ap);
+			info(ap->logopt, "verbose logging set for %s", ap->path);
+		}
+	} else {
+		if (ap->logopt & LOGOPT_ANY)
+			info(ap->logopt, "basic logging set for %s", ap->path);
+		set_log_norm_ap(ap);
+	}
+}
+
+static int set_log_priority(const char *path, int priority)
+{
+	struct ioctl_ops *ops = get_ioctl_ops();
+	char buf[FIFO_BUF_SIZE];
+	dev_t devid;
+	int fd;
+	int ret;
+
+	if (!ops) {
+		fprintf(stderr, "Could not get ioctl ops\n");
+		return -1;
+	} else {
+		ret = ops->mount_device(LOGOPT_ANY, path, 0, &devid);
+		if (ret == -1 || ret == 0) {
+			fprintf(stderr,
+				"Could not find device id for mount %s\n", path);
+			return -1;
+		}
+	}
+
+	if (priority > LOG_DEBUG || priority < LOG_EMERG) {
+		fprintf(stderr, "Log priority %d is invalid.\n", priority);
+		fprintf(stderr, "Please specify a number in the range 0-7.\n");
+		return -1;
+	}
+
+	/*
+	 * This is an ascii based protocol, so we want the string
+	 * representation of the integer log priority.
+	 */
+	ret = snprintf(buf, sizeof(buf), "%ld %d", devid, priority);
+	if (ret >= FIFO_BUF_SIZE) {
+		fprintf(stderr, "Invalid device id or log priotity\n");
+		return -1;
+	}
+
+	/*
+	 * Specify O_NONBLOCK so that the open will fail if there is no
+	 * daemon reading from the other side of the FIFO.
+	 */
+	fd = open_fd(cmd_pipe_name, O_WRONLY|O_NONBLOCK);
+	if (fd < 0) {
+		fprintf(stderr, "%s: open of %s failed with %s\n",
+			__FUNCTION__, cmd_pipe_name, strerror(errno));
+		fprintf(stderr, "%s: perhaps the fifo wasn't setup,"
+			" please check your log for more information\n", __FUNCTION__);
+		return -1;
+	}
+
+	if (write(fd, buf, sizeof(buf)) != sizeof(buf)) {
+		fprintf(stderr, "Failed to change logging priority.  ");
+		fprintf(stderr, "write to fifo failed: %s.\n",
+			strerror(errno));
+		close(fd);
+		return -1;
+	}
+	close(fd);
+	fprintf(stdout, "Successfully set log priority for %s.\n", path);
+
+	return 0;
+}
+
+static void cmd_pipe_dummy(int sig)
+{
+}
+
+static void *cmd_pipe_handler(void *arg)
+{
+	struct sigaction sa;
+	sigset_t signalset;
+	struct pollfd fds[1];
+	int pollfds = 1;
+	char buf[MAX_ERR_BUF];
+	char *estr;
+
+	if (create_cmd_pipe_fifo())
+		return NULL;
+
+	fds[0].fd = cmd_pipe_fifo;
+	fds[0].events = POLLIN;
+
+	sa.sa_handler = cmd_pipe_dummy;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = 0;
+	if (sigaction(SIGPIPE, &sa, NULL) == -1) {
+		error(LOGOPT_ANY, "failed to set signal handler %d", errno);
+		return NULL;
+	}
+
+	sigfillset(&signalset);
+	sigdelset(&signalset, SIGPIPE);
+
+	while (1) {
+		cmd_pipe_mutex_lock();
+		if (done)
+			break;
+		cmd_pipe_mutex_unlock();
+
+		errno = 0;
+		if (ppoll(fds, pollfds, NULL, &signalset) == -1) {
+			if (errno == EINTR)
+				continue;
+			cmd_pipe_mutex_lock();
+			estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			logerr("poll failed: %s", estr);
+			break;
+		}
+
+		if (fds[0].revents & POLLIN) {
+			debug(LOGOPT_ANY, "message pending on control fifo.");
+			handle_cmd_pipe_fifo_message(fds[0].fd);
+		}
+	}
+	destroy_cmd_pipe_fifo();
+	cmd_pipe_mutex_unlock();
+	return NULL;
+}
+
+int start_cmd_pipe_handler(void)
+{
+	pthread_t thid;
+	pthread_attr_t attrs;
+	pthread_attr_t *pattrs = &attrs;
+	int status;
+
+	status = pthread_attr_init(pattrs);
+	if (status)
+		pattrs = NULL;
+	else
+		pthread_attr_setdetachstate(pattrs, PTHREAD_CREATE_DETACHED);
+
+	status = pthread_create(&thid, pattrs, cmd_pipe_handler, NULL);
+
+	if (pattrs)
+		pthread_attr_destroy(pattrs);
+
+	if (!status)
+		cmd_pipe_thid = thid;
+
+	return !status;
+}
+
+void finish_cmd_pipe_handler(void)
+{
+	cmd_pipe_mutex_lock();
+	if (cmd_pipe_thid == -1 || done)
+	       goto exit;
+	done = 1;
+	pthread_kill(cmd_pipe_thid, SIGPIPE);
+exit:
+	cmd_pipe_mutex_unlock();
+}
+
 static void return_start_status(void *arg)
 {
 	struct startup_cond *sc;
@@ -1732,9 +1771,6 @@ static void handle_mounts_cleanup(void *arg)
 	if (submount) {
 		struct mnt_list *mnt;
 
-		/* We are finishing up */
-		ap->parent->submnt_count--;
-
 		/* Submount at ap->path belongs to parent submount list. */
 		mnts_remove_submount(ap->path);
 		/* Also remove from parent mounted list */
@@ -1752,19 +1788,6 @@ static void handle_mounts_cleanup(void *arg)
 
 	info(logopt, "shut down path %s", ap->path);
 
-	destroy_logpri_fifo(ap);
-
-	/*
-	 * Submounts are detached threads and don't belong to the
-	 * master map entry list so we need to free their resources
-	 * here.
-	 */
-	if (submount) {
-		master_source_unlock(ap->parent->entry);
-		master_free_mapent_sources(ap->entry, 1);
-		master_free_mapent(ap->entry);
-	}
-
 	if (clean) {
 		if (rmdir(ap->path) == -1) {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
@@ -1777,48 +1800,17 @@ static void handle_mounts_cleanup(void *arg)
 	master_source_unlock(ap->entry);
 
 	/*
-	 * If we are not a submount send a signal to the signal handler
-	 * so it can join with any completed handle_mounts() threads and
-	 * perform final cleanup.
+	 * Send a signal to the signal handler so it can join with any
+	 * completed handle_mounts() threads and perform final cleanup.
 	 */
-	if (!submount && !pending)
-		pthread_kill(state_mach_thid, SIGTERM);
+	if (!pending)
+		pthread_kill(signal_handler_thid, SIGTERM);
 
 	master_mutex_unlock();
 
 	return;
 }
 
-static int submount_source_writelock_nested(struct autofs_point *ap)
-{
-	struct autofs_point *parent = ap->parent;
-	int status;
-
-	status = pthread_rwlock_trywrlock(&parent->entry->source_lock);
-	if (status)
-		goto done;
-
-	status = pthread_rwlock_trywrlock(&ap->entry->source_lock);
-	if (status)
-		master_source_unlock(parent->entry);
-
-done:
-	if (status && status != EBUSY) {
-		logmsg("submount nested master_mapent source write lock failed");
-		fatal(status);
-	}
-
-	return status;
-}
-
-static void submount_source_unlock_nested(struct autofs_point *ap)
-{
-	struct autofs_point *parent = ap->parent;
-
-	master_source_unlock(ap->entry);
-	master_source_unlock(parent->entry);
-}
-
 int handle_mounts_exit(struct autofs_point *ap)
 {
 	int ret, cur_state;
@@ -1833,33 +1825,13 @@ int handle_mounts_exit(struct autofs_point *ap)
 
 	master_mutex_lock();
 
-	if (!ap->submount)
-		master_source_writelock(ap->entry);
-	else {
-		/*
-		 * If a mount request arrives before the locks are
-		 * aquired just return to ready state.
-		 */
-		ret = submount_source_writelock_nested(ap);
-		if (ret) {
-			warn(ap->logopt,
-			     "can't shutdown submount: mount in progress");
-			/* Return to ST_READY is done immediately */
-			st_add_task(ap, ST_READY);
-			master_mutex_unlock();
-			pthread_setcancelstate(cur_state, NULL);
-			return 0;
-		}
-	}
+	master_source_writelock(ap->entry);
 
 	if (ap->state != ST_SHUTDOWN) {
 		conditional_alarm_add(ap, ap->exp_runfreq);
 		/* Return to ST_READY is done immediately */
 		st_add_task(ap, ST_READY);
-		if (ap->submount)
-			submount_source_unlock_nested(ap);
-		else
-			master_source_unlock(ap->entry);
+		master_source_unlock(ap->entry);
 		master_mutex_unlock();
 
 		pthread_setcancelstate(cur_state, NULL);
@@ -1877,7 +1849,7 @@ int handle_mounts_exit(struct autofs_point *ap)
 	 * to check for possible recovery.
 	 */
 	if (ap->type == LKP_DIRECT) {
-		umount_autofs(ap, NULL);
+		umount_autofs(ap);
 		handle_mounts_cleanup(ap);
 		return 1;
 	}
@@ -1888,7 +1860,7 @@ int handle_mounts_exit(struct autofs_point *ap)
 	 * so we can continue. This can happen if a lookup
 	 * occurs while we're trying to umount.
 	 */
-	ret = umount_autofs(ap, NULL);
+	ret = umount_autofs(ap);
 	if (!ret) {
 		set_indirect_mount_tree_catatonic(ap);
 		handle_mounts_cleanup(ap);
@@ -1900,10 +1872,7 @@ int handle_mounts_exit(struct autofs_point *ap)
 	conditional_alarm_add(ap, ap->exp_runfreq);
 	/* Return to ST_READY is done immediately */
 	st_add_task(ap, ST_READY);
-	if (ap->submount)
-		submount_source_unlock_nested(ap);
-	else
-		master_source_unlock(ap->entry);
+	master_source_unlock(ap->entry);
 	master_mutex_unlock();
 
 	pthread_setcancelstate(cur_state, NULL);
@@ -1916,12 +1885,10 @@ void *handle_mounts(void *arg)
 	struct startup_cond *suc;
 	struct autofs_point *ap;
 	int cancel_state, status = 0;
-	char *root;
 
 	suc = (struct startup_cond *) arg;
 
 	ap = suc->ap;
-	root = strdup(suc->root);
 
 	pthread_cleanup_push(return_start_status, suc);
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel_state);
@@ -1929,30 +1896,18 @@ void *handle_mounts(void *arg)
 	status = pthread_mutex_lock(&suc->mutex);
 	if (status) {
 		logerr("failed to lock startup condition mutex!");
-		if (root)
-			free(root);
 		fatal(status);
 	}
 
-	if (!root) {
-		crit(ap->logopt, "failed to alloc string root");
-		suc->status = 1;
-		pthread_setcancelstate(cancel_state, NULL);
-		pthread_exit(NULL);
-	}
-
-	if (mount_autofs(ap, root) < 0) {
+	if (mount_autofs(ap) < 0) {
 		if (!(do_force_unlink & UNLINK_AND_EXIT))
 			crit(ap->logopt, "mount of %s failed!", ap->path);
 		suc->status = 1;
-		umount_autofs(ap, root);
-		free(root);
+		umount_autofs(ap);
 		pthread_setcancelstate(cancel_state, NULL);
 		pthread_exit(NULL);
 	}
 
-	free(root);
-
 	if (ap->flags & MOUNT_FLAG_NOBIND)
 		info(ap->logopt, "bind mounts disabled");
 
@@ -2011,7 +1966,9 @@ static void usage(void)
 	        "			maximum wait time (seconds) for master\n"
 	        "			map to become available\n"
 		"	-v --verbose	be verbose\n"
-		"	-d --debug	log debuging info\n"
+		"	-d[level]\n"
+		"	--debug[=level]\n"
+		"			log debugging info\n"
 		"	-Dvariable=value, --define variable=value\n"
 		"			define global macro variable\n"
 		"	-S --systemd-service\n"
@@ -2023,6 +1980,8 @@ static void usage(void)
 		"			dump automounter maps and exit\n"
 		"	-n --negative-timeout n\n"
 		"			set the timeout for failed key lookups.\n"
+		"	-P --positive-timeout n\n"
+		"			set the positive timeout for key lookups.\n"
 		"	-O --global-options\n"
 		"			specify global mount options\n"
 		"	-l --set-log-priority priority path [path,...]\n"
@@ -2228,7 +2187,7 @@ static void do_master_list_reset(struct master *master)
 
 static int do_master_read_master(struct master *master, time_t *age, int wait)
 {
-	sigset_t signalset;
+	sigset_t signalset, savesigset;
 	/* Wait must be at least 1 second */
 	unsigned int retry_wait = 2;
 	unsigned int elapsed = 0;
@@ -2239,7 +2198,7 @@ static int do_master_read_master(struct master *master, time_t *age, int wait)
 	sigaddset(&signalset, SIGTERM);
 	sigaddset(&signalset, SIGINT);
 	sigaddset(&signalset, SIGHUP);
-	sigprocmask(SIG_UNBLOCK, &signalset, NULL);
+	pthread_sigmask(SIG_UNBLOCK, &signalset, &savesigset);
 
 	while (1) {
 		struct timespec t = { retry_wait, 0 };
@@ -2265,7 +2224,7 @@ static int do_master_read_master(struct master *master, time_t *age, int wait)
 		}
 	}
 
-	sigprocmask(SIG_BLOCK, &signalset, NULL);
+	pthread_sigmask(SIG_SETMASK, &savesigset, NULL);
 
 	return ret;
 }
@@ -2276,23 +2235,26 @@ int main(int argc, char *argv[])
 	int logpri = -1;
 	unsigned int flags;
 	unsigned int logging;
+	int debug_level = 0;
 	unsigned master_read;
 	int master_wait;
 	time_t timeout;
 	time_t age = monotonic_time(NULL);
 	struct rlimit rlim;
-	const char *options = "+hp:t:vmdD:SfVrO:l:n:CFUM:";
+	unsigned long max_open_files;
+	const char *options = "+hp:t:vmd::D:SfVrO:l:n:P:CFUM:";
 	static const struct option long_options[] = {
 		{"help", 0, 0, 'h'},
 		{"pid-file", 1, 0, 'p'},
 		{"timeout", 1, 0, 't'},
 		{"verbose", 0, 0, 'v'},
-		{"debug", 0, 0, 'd'},
+		{"debug", 2, 0, 'd'},
 		{"define", 1, 0, 'D'},
 		{"systemd-service", 0, 0, 'S'},
 		{"foreground", 0, 0, 'f'},
 		{"random-multimount-selection", 0, 0, 'r'},
 		{"negative-timeout", 1, 0, 'n'},
+		{"positive-timeout", 1, 0, 'P'},
 		{"dumpmaps", 0, 0, 'm'},
 		{"global-options", 1, 0, 'O'},
 		{"version", 0, 0, 'V'},
@@ -2312,7 +2274,7 @@ int main(int argc, char *argv[])
 	sigdelset(&block_sigs, SIGILL);
 	sigdelset(&block_sigs, SIGFPE);
 	sigdelset(&block_sigs, SIGTRAP);
-	sigprocmask(SIG_BLOCK, &block_sigs, NULL);
+	pthread_sigmask(SIG_BLOCK, &block_sigs, NULL);
 
 	program = argv[0];
 
@@ -2357,6 +2319,8 @@ int main(int argc, char *argv[])
 
 		case 'd':
 			logging |= LOGOPT_DEBUG;
+			if (optarg)
+				debug_level = getsnumopt(optarg, opt);
 			break;
 
 		case 'D':
@@ -2383,6 +2347,10 @@ int main(int argc, char *argv[])
 			global_negative_timeout = getnumopt(optarg, opt);
 			break;
 
+		case 'P':
+			global_positive_timeout = getnumopt(optarg, opt);
+			break;
+
 		case 'm':
 			flags |= DAEMON_FLAGS_DUMP_MAPS;
 			break;
@@ -2433,7 +2401,8 @@ int main(int argc, char *argv[])
 
 		case '?':
 		case ':':
-			printf("%s: Ambiguous or unknown options\n", program);
+			fprintf(stderr, "%s: Ambiguous or unknown options\n", program);
+			fprintf(stderr, "Try `%s --help` for more information\n", program);
 			exit(1);
 		}
 	}
@@ -2442,7 +2411,7 @@ int main(int argc, char *argv[])
 		set_log_verbose();
 
 	if (logging & LOGOPT_DEBUG)
-		set_log_debug();
+		set_log_debug(debug_level);
 
 	if (geteuid() != 0) {
 		fprintf(stderr, "%s: this program must be run by root.\n",
@@ -2485,10 +2454,13 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
+	max_open_files = defaults_get_open_file_limit();
+
 	res = getrlimit(RLIMIT_NOFILE, &rlim);
-	if (res == -1 || rlim.rlim_max <= MAX_OPEN_FILES)  {
-		rlim.rlim_cur = MAX_OPEN_FILES;
-		rlim.rlim_max = MAX_OPEN_FILES;
+	if (res == -1 || rlim.rlim_cur <= max_open_files)  {
+		rlim.rlim_cur = max_open_files;
+		if (rlim.rlim_max < max_open_files)
+			rlim.rlim_max = max_open_files;
 	}
 	res = setrlimit(RLIMIT_NOFILE, &rlim);
 	if (res)
@@ -2524,6 +2496,23 @@ int main(int argc, char *argv[])
 				master = argv[2];
 		}
 
+		status = pthread_key_create(&key_thread_stdenv_vars,
+					key_thread_stdenv_vars_destroy);
+		if (status) {
+			logerr("%s: failed to create thread data key for std env vars!",
+			       program);
+			macro_free_global_table();
+			exit(1);
+		}
+
+		status = pthread_key_create(&key_thread_attempt_id, free);
+		if (status) {
+			logerr("%s: failed to create thread data key for attempt ID!",
+			       program);
+			macro_free_global_table();
+			exit(1);
+		}
+
 		if (master)
 			master_list = master_new(NULL, timeout, flags);
 		else
@@ -2658,6 +2647,18 @@ int main(int argc, char *argv[])
 
 	init_ioctl_ctl();
 
+	if (!start_cmd_pipe_handler()) {
+		logerr("%s: failed to create command pipe handler thread!", program);
+		master_kill(master_list);
+		if (start_pipefd[1] != -1) {
+			res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+			close(start_pipefd[1]);
+		}
+		release_flag_file();
+		macro_free_global_table();
+		exit(1);
+	}
+
 	if (!alarm_start_handler()) {
 		logerr("%s: failed to create alarm handler thread!", program);
 		master_kill(master_list);
@@ -2723,13 +2724,7 @@ int main(int argc, char *argv[])
 		}
 	}
 
-	/* If the option to unlink all autofs mounts and exit has
-	 * been given remove logpri fifo pipe files as all the mounts
-	 * will be detached leaving them stale.
-	 */
-	if (do_force_unlink & UNLINK_AND_EXIT)
-		cleanup_stale_logpri_fifo_pipes();
-	else {
+	if (!(do_force_unlink & UNLINK_AND_EXIT)) {
 		/*
 		 * Mmm ... reset force unlink umount so we don't also do
 		 * this in future when we receive a HUP signal.
@@ -2747,12 +2742,14 @@ int main(int argc, char *argv[])
 			sd_notify(1, "READY=1");
 #endif
 
-		state_mach_thid = pthread_self();
-		statemachine(NULL);
+		signal_handler_thid = pthread_self();
+		signal_handler(NULL);
 	}
 
 	master_kill(master_list);
 
+	finish_cmd_pipe_handler();
+
 	if (pid_file) {
 		unlink(pid_file);
 		pid_file = NULL;
