@@ -40,7 +40,7 @@
 /* Attribute to create detached thread */
 extern pthread_attr_t th_attr_detached;
 
-static int do_mount_autofs_indirect(struct autofs_point *ap, const char *root)
+static int do_mount_autofs_indirect(struct autofs_point *ap)
 {
 	const char *str_indirect = mount_type_str(t_indirect);
 	struct ioctl_ops *ops = get_ioctl_ops();
@@ -89,11 +89,11 @@ static int do_mount_autofs_indirect(struct autofs_point *ap, const char *root)
 	}
 
 	/* In case the directory doesn't exist, try to mkdir it */
-	if (mkdir_path(root, mp_mode) < 0) {
+	if (mkdir_path(ap->path, mp_mode) < 0) {
 		if (errno != EEXIST && errno != EROFS) {
 			crit(ap->logopt,
 			     "failed to create autofs directory %s",
-			     root);
+			     ap->path);
 			goto out_err;
 		}
 		/* If we recieve an error, and it's EEXIST or EROFS we know
@@ -108,59 +108,54 @@ static int do_mount_autofs_indirect(struct autofs_point *ap, const char *root)
 	if (!type || strcmp(ap->entry->maps->type, "hosts"))
 		map_name = ap->entry->maps->argv[0];
 
-	ret = mount(map_name, root, "autofs", MS_MGC_VAL, options);
+	ret = mount(map_name, ap->path, "autofs", MS_MGC_VAL, options);
 	if (ret) {
 		crit(ap->logopt,
-		     "failed to mount autofs path %s at %s", ap->path, root);
+		     "failed to mount autofs at %s", ap->path);
 		goto out_rmdir;
 	}
 
 	free(options);
 	options = NULL;
 
-	ret = stat(root, &st);
+	ret = stat(ap->path, &st);
 	if (ret == -1) {
 		crit(ap->logopt,
 		     "failed to stat mount for autofs path %s", ap->path);
 		goto out_umount;
 	}
+	ap->dev = st.st_dev;	/* Device number for mount point checks */
 
-	if (ap->mode && (err = chmod(root, ap->mode)))
+	if (ap->mode && (err = chmod(ap->path, ap->mode)))
 		warn(ap->logopt, "failed to change mode of %s", ap->path);
 
-	if (ops->open(ap->logopt, &ap->ioctlfd, st.st_dev, root)) {
+	ap->ioctlfd = open_ioctlfd(ap, ap->path, ap->dev);
+	if (ap->ioctlfd == -1) {
 		crit(ap->logopt,
 		     "failed to create ioctl fd for autofs path %s", ap->path);
 		goto out_umount;
 	}
 
-	ap->dev = st.st_dev;	/* Device number for mount point checks */
-
 	ops->timeout(ap->logopt, ap->ioctlfd, timeout);
-	if (ap->logopt & LOGOPT_DEBUG)
-		notify_mount_result(ap, root, timeout, str_indirect);
-	else
-		notify_mount_result(ap, ap->path, timeout, str_indirect);
+	notify_mount_result(ap, ap->path, timeout, str_indirect);
 
 	return 0;
 
 out_umount:
-	umount(root);
+	umount(ap->path);
 out_rmdir:
 	if (ap->flags & MOUNT_FLAG_DIR_CREATED)
-		rmdir(root);
+		rmdir(ap->path);
 out_err:
 	if (options)
 		free(options);
-	close(ap->state_pipe[0]);
-	close(ap->state_pipe[1]);
 	close(ap->pipefd);
 	close(ap->kpipefd);
 
 	return -1;
 }
 
-int mount_autofs_indirect(struct autofs_point *ap, const char *root)
+int mount_autofs_indirect(struct autofs_point *ap)
 {
 	time_t now = monotonic_time(NULL);
 	int status;
@@ -182,11 +177,11 @@ int mount_autofs_indirect(struct autofs_point *ap, const char *root)
 		}
 	}
 
-	status = do_mount_autofs_indirect(ap, root);
+	status = do_mount_autofs_indirect(ap);
 	if (status < 0)
 		return -1;
 
-	map = lookup_ghost(ap, root);
+	map = lookup_ghost(ap);
 	if (map & LKP_FAIL) {
 		if (map & LKP_DIRECT) {
 			error(ap->logopt,
@@ -216,11 +211,6 @@ void close_mount_fds(struct autofs_point *ap)
 	if (ap->submount)
 		lookup_source_close_ioctlfd(ap->parent, ap->path);
 
-	close(ap->state_pipe[0]);
-	close(ap->state_pipe[1]);
-	ap->state_pipe[0] = -1;
-	ap->state_pipe[1] = -1;
-
 	if (ap->pipefd >= 0)
 		close(ap->pipefd);
 
@@ -230,28 +220,13 @@ void close_mount_fds(struct autofs_point *ap)
 	return;
 }
 
-int umount_autofs_indirect(struct autofs_point *ap, const char *root)
+int umount_autofs_indirect(struct autofs_point *ap)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	char buf[MAX_ERR_BUF];
-	char mountpoint[PATH_MAX + 1];
 	int rv, retries;
 	unsigned int unused;
 
-	if (root) {
-		if (strlen(root) > PATH_MAX) {
-			error(ap->logopt, "mountpoint path too long");
-			return 1;
-		}
-		strcpy(mountpoint, root);
-	} else {
-		if (ap->len > PATH_MAX) {
-			error(ap->logopt, "mountpoint path too long");
-			return 1;
-		}
-		strcpy(mountpoint, ap->path);
-	}
-
 	/* If we are trying to shutdown make sure we can umount */
 	rv = ops->askumount(ap->logopt, ap->ioctlfd, &unused);
 	if (rv == -1) {
@@ -273,7 +248,7 @@ int umount_autofs_indirect(struct autofs_point *ap, const char *root)
 	sched_yield();
 
 	retries = UMOUNT_RETRIES;
-	while ((rv = umount(mountpoint)) == -1 && retries--) {
+	while ((rv = umount(ap->path)) == -1 && retries--) {
 		struct timespec tm = {0, 50000000};
 		if (errno != EBUSY)
 			break;
@@ -285,13 +260,13 @@ int umount_autofs_indirect(struct autofs_point *ap, const char *root)
 		case ENOENT:
 		case EINVAL:
 			error(ap->logopt,
-			      "mount point %s does not exist", mountpoint);
+			      "mount point %s does not exist", ap->path);
 			close_mount_fds(ap);
 			return 0;
 			break;
 		case EBUSY:
 			debug(ap->logopt,
-			      "mount point %s is in use", mountpoint);
+			      "mount point %s is in use", ap->path);
 			if (ap->state == ST_SHUTDOWN_FORCE) {
 				close_mount_fds(ap);
 				goto force_umount;
@@ -309,12 +284,11 @@ int umount_autofs_indirect(struct autofs_point *ap, const char *root)
 					return 0;
 				}
 #endif
-				ops->open(ap->logopt,
-					  &ap->ioctlfd, ap->dev, mountpoint);
+				ap->ioctlfd = open_ioctlfd(ap, ap->path, ap->dev);
 				if (ap->ioctlfd < 0) {
 					warn(ap->logopt,
 					     "could not recover autofs path %s",
-					     mountpoint);
+					     ap->path);
 					close_mount_fds(ap);
 					return 0;
 				}
@@ -339,12 +313,12 @@ int umount_autofs_indirect(struct autofs_point *ap, const char *root)
 force_umount:
 	if (rv != 0) {
 		warn(ap->logopt,
-		     "forcing umount of indirect mount %s", mountpoint);
-		rv = umount2(mountpoint, MNT_DETACH);
+		     "forcing umount of indirect mount %s", ap->path);
+		rv = umount2(ap->path, MNT_DETACH);
 	} else {
-		info(ap->logopt, "umounted indirect mount %s", mountpoint);
+		info(ap->logopt, "umounting indirect mount %s succeeded", ap->path);
 		if (ap->submount)
-			rm_unwanted(ap, mountpoint, 1);
+			rm_unwanted(ap, ap->path, 1);
 	}
 
 	return rv;
@@ -369,6 +343,7 @@ void *expire_proc_indirect(void *arg)
 	int offsets, submnts, count;
 	int ioctlfd, cur_state;
 	int status, ret, left;
+	int retries;
 
 	ea = (struct expire_args *) arg;
 
@@ -408,8 +383,11 @@ void *expire_proc_indirect(void *arg)
 			 * one of them and pass on the state change.
 			 */
 			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-			if (mnt->flags & MNTS_AUTOFS)
+			if (mnt->flags & MNTS_AUTOFS) {
 				master_notify_submount(ap, mnt->mp, ap->state);
+				pthread_setcancelstate(cur_state, NULL);
+				continue;
+			}
 
 			/* An offset without a real mount, check for manual umount */
 			if (mnt->flags & MNTS_OFFSET &&
@@ -513,9 +491,19 @@ void *expire_proc_indirect(void *arg)
 	 * If there are no more real mounts left we could still
 	 * have some offset mounts with no '/' offset or symlinks
 	 * so we need to umount or unlink them here.
+	 *
+	 * The dentry info last_used field is set to 'now' when a
+	 * dentry is selected for expire so that it isn't immediately
+	 * selected again if the expire fails. But this can't work
+	 * for immediate expires so the count_mounts() function must
+	 * be used to limit the number of expire iterations.
 	 */
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-	while (1) {
+	if (how == AUTOFS_EXP_IMMEDIATE)
+		retries = count_mounts(ap, ap->path, ap->dev);
+	else
+		retries = -1;
+	while (retries--) {
 		ret = ops->expire(ap->logopt, ap->ioctlfd, ap->path, how);
 		if (ret != 0 && errno == EAGAIN)
 			break;
@@ -683,13 +671,18 @@ int handle_packet_expire_indirect(struct autofs_point *ap, autofs_packet_expire_
 	return 0;
 }
 
-static void mount_send_fail(void *arg)
+static void mount_send_status(void *arg)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *mt = arg;
 	struct autofs_point *ap = mt->ap;
-	ops->send_fail(ap->logopt,
-		       ap->ioctlfd, mt->wait_queue_token, -ENOENT);
+
+	if (mt->status)
+		ops->send_fail(ap->logopt, ap->ioctlfd,
+			       mt->wait_queue_token, mt->status);
+	else
+		ops->send_ready(ap->logopt,
+				ap->ioctlfd, mt->wait_queue_token);
 }
 
 static void *do_mount_indirect(void *arg)
@@ -718,7 +711,8 @@ static void *do_mount_indirect(void *arg)
 
 	pending_mutex_unlock(args);
 
-	pthread_cleanup_push(mount_send_fail, &mt);
+	mt.status = 0;
+	pthread_cleanup_push(mount_send_status, &mt);
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 
@@ -731,9 +725,7 @@ static void *do_mount_indirect(void *arg)
 	len = ncat_path(buf, sizeof(buf), ap->path, mt.name, mt.len);
 	if (!len) {
 		crit(ap->logopt, "path to be mounted is to long");
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, mt.wait_queue_token,
-			      -ENAMETOOLONG);
+		mt.status = -ENAMETOOLONG;
 		pthread_setcancelstate(state, NULL);
 		pthread_exit(NULL);
 	}
@@ -742,7 +734,6 @@ static void *do_mount_indirect(void *arg)
 	if (status != -1 && !(S_ISDIR(st.st_mode) && st.st_dev == mt.dev)) {
 		error(ap->logopt,
 		      "indirect trigger not valid or already mounted %s", buf);
-		ops->send_ready(ap->logopt, ap->ioctlfd, mt.wait_queue_token);
 		pthread_setcancelstate(state, NULL);
 		pthread_exit(NULL);
 	}
@@ -805,9 +796,9 @@ int handle_packet_missing_indirect(struct autofs_point *ap, autofs_packet_missin
 		return 0;
 	}
 
-	/* Check if we recorded a mount fail for this key anywhere */
 	me = lookup_source_mapent(ap, pkt->name, LKP_DISTINCT);
 	if (me) {
+		/* Check if we recorded a mount fail for this key */
 		if (me->status >= monotonic_time(NULL)) {
 			ops->send_fail(ap->logopt, ap->ioctlfd,
 				       pkt->wait_queue_token, -ENOENT);
@@ -816,6 +807,16 @@ int handle_packet_missing_indirect(struct autofs_point *ap, autofs_packet_missin
 			pthread_setcancelstate(state, NULL);
 			return 0;
 		}
+
+		/* Ignore nulled indirect map entries */
+		if (starts_with_null_opt(me->mapent)) {
+			ops->send_fail(ap->logopt, ap->ioctlfd,
+				       pkt->wait_queue_token, -ENOENT);
+			cache_unlock(me->mc);
+			master_mutex_unlock();
+			pthread_setcancelstate(state, NULL);
+			return 0;
+		}
 		cache_unlock(me->mc);
 	}
 
