@@ -25,6 +25,8 @@
 #include "automount.h"
 #include "nsswitch.h"
 
+extern long global_positive_timeout;
+
 static void nsslist_cleanup(void *arg)
 {
 	struct list_head *nsslist = (struct list_head *) arg;
@@ -169,7 +171,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 	/* If it starts with a '/' it has to be a file or LDAP map */
 	if (*master->name == '/') {
 		if (*(master->name + 1) == '/') {
-			debug(logopt, "reading master ldap %s", master->name);
+			info(logopt, "reading ldap master %s", master->name);
 			result = do_read_master(master, "ldap", age);
 		} else {
 			debug(logopt, "reading master file %s", master->name);
@@ -215,7 +217,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 				 */
 				if (strncmp(name, "ldap", 4)) {
 					master->name = tmp + 1;
-					debug(logopt, "reading master %s %s",
+					info(logopt, "reading %s master %s",
 					      source, master->name);
 				} else {
 					master->name = name;
@@ -260,8 +262,8 @@ int lookup_nss_read_master(struct master *master, time_t age)
 		    strncmp(this->source, "sss", 3))
 			continue;
 
-		debug(logopt,
-		      "reading master %s %s", this->source, master->name);
+		info(logopt,
+		      "reading %s master %s", this->source, master->name);
 
 		result = read_master_map(master, this->source, age);
 
@@ -318,36 +320,54 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 	struct lookup_mod *lookup;
 	int status;
 
-	lookup = NULL;
-	master_source_writelock(ap->entry);
-	if (!map->lookup) {
-		status = open_lookup(map->type, "", map->format,
-				     map->argc, map->argv, &lookup);
-		if (status != NSS_STATUS_SUCCESS) {
-			master_source_unlock(ap->entry);
-			debug(ap->logopt,
-			      "lookup module %s open failed", map->type);
-			return status;
-		}
-		map->lookup = lookup;
-	} else {
-		lookup = map->lookup;
-		status = lookup->lookup_reinit(map->format,
-					       map->argc, map->argv,
-					       &lookup->context);
-		if (status)
-			warn(ap->logopt,
-			     "lookup module %s reinit failed", map->type);
-	}
-	master_source_unlock(ap->entry);
-
 	if (!map->stale)
 		return NSS_STATUS_SUCCESS;
 
-	master_source_current_wait(ap->entry);
-	ap->entry->current = map;
+	/* If this readmap is the result of trying to mount a submount
+	 * the readlock may already be held if the map is the same as
+	 * that of the caller. In that case the map has already been
+	 * read so just skip the map open/reinit.
+	 */
+	status = map_module_try_writelock(map);
+	if (status) {
+		if (!map->lookup) {
+			error(ap->logopt, "map module lock not held as expected");
+			return NSS_STATUS_UNAVAIL;
+		}
+	} else {
+		if (!map->lookup) {
+			pthread_cleanup_push(map_module_lock_cleanup, map);
+			status = open_lookup(map->type, "", map->format,
+					     map->argc, map->argv, &lookup);
+			pthread_cleanup_pop(0);
+			if (status != NSS_STATUS_SUCCESS) {
+				map_module_unlock(map);
+				debug(ap->logopt,
+				      "lookup module %s open failed", map->type);
+				return status;
+			}
+			map->lookup = lookup;
+		} else {
+			pthread_cleanup_push(map_module_lock_cleanup, map);
+			status = map->lookup->lookup_reinit(map->format,
+							    map->argc, map->argv,
+							    &map->lookup->context);
+			pthread_cleanup_pop(0);
+			if (status) {
+				map_module_unlock(map);
+				warn(ap->logopt,
+				     "lookup module %s reinit failed", map->type);
+				return status;
+			}
+		}
+		map_module_unlock(map);
+	}
 
-	status = lookup->lookup_read_map(ap, age, lookup->context);
+	pthread_cleanup_push(map_module_lock_cleanup, map);
+	map_module_readlock(map);
+	lookup = map->lookup;
+	status = lookup->lookup_read_map(ap, map, age, lookup->context);
+	pthread_cleanup_pop(1);
 
 	if (status != NSS_STATUS_SUCCESS)
 		map->stale = 0;
@@ -397,7 +417,7 @@ static int read_file_source_instance(struct autofs_point *ap, struct map_source
 		return NSS_STATUS_NOTFOUND;
 	}
 
-	if (st.st_mode & __S_IEXEC)
+	if (st.st_mode & S_IEXEC)
 		type = src_prog;
 	else
 		type = src_file;
@@ -549,8 +569,8 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 			if (!strncmp(map->type, "multi", 5))
 				debug(ap->logopt, "reading multi map");
 			else
-				debug(ap->logopt,
-				      "reading map %s %s",
+				info(ap->logopt,
+				      "reading %s map %s",
 				       map->type, map->argv[0]);
 			result = lookup_map_read_map(ap, map, age);
 			map = map->next;
@@ -566,8 +586,8 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 					continue;
 				}
 				map->type = tmp;
-				debug(ap->logopt,
-				      "reading map %s %s", tmp, map->argv[0]);
+				info(ap->logopt,
+				      "reading %s map %s", tmp, map->argv[0]);
 				result = do_read_map(ap, map, age);
 			} else {
 				debug(ap->logopt,
@@ -602,8 +622,8 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 				continue;
 			}
 
-			debug(ap->logopt,
-			      "reading map %s %s", this->source, map->argv[0]);
+			info(ap->logopt,
+			      "reading %s map %s", this->source, map->argv[0]);
 
 			result = read_map_source(this, ap, map, age);
 			if (result == NSS_STATUS_UNKNOWN)
@@ -691,7 +711,7 @@ static char *make_browse_path(unsigned int logopt,
 	return path;
 }
 
-int lookup_ghost(struct autofs_point *ap, const char *root)
+int lookup_ghost(struct autofs_point *ap)
 {
 	struct master_mapent *entry = ap->entry;
 	struct map_source *map;
@@ -754,30 +774,22 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 				goto next;
 			}
 
+			/* Ignore nulled indirect map entries */
+			if (starts_with_null_opt(me->mapent))
+				goto next;
+
 			fullpath = make_browse_path(ap->logopt,
-						    root, me->key, ap->pref);
+						    ap->path, me->key, ap->pref);
 			if (!fullpath)
 				goto next;
 
-			ret = stat(fullpath, &st);
-			if (ret == -1 && errno != ENOENT) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				warn(ap->logopt, "stat error %s", estr);
-				free(fullpath);
-				goto next;
-			}
-
-			/* Directory already exists? */
-			if (!ret) {
-				free(fullpath);
-				goto next;
-			}
-
 			ret = mkdir_path(fullpath, mp_mode);
-			if (ret < 0 && errno != EEXIST) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				warn(ap->logopt,
-				     "mkdir_path %s failed: %s", fullpath, estr);
+			if (ret < 0) {
+				if (errno != EEXIST) {
+					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+					warn(ap->logopt,
+					     "mkdir_path %s failed: %s", fullpath, estr);
+				}
 				free(fullpath);
 				goto next;
 			}
@@ -805,22 +817,25 @@ int do_lookup_mount(struct autofs_point *ap, struct map_source *map, const char
 	int status;
 
 	if (!map->lookup) {
-		status = open_lookup(map->type, "",
-				     map->format, map->argc, map->argv, &lookup);
-		if (status != NSS_STATUS_SUCCESS) {
-			debug(ap->logopt,
-			      "lookup module %s open failed", map->type);
-			return status;
+		map_module_writelock(map);
+		if (!map->lookup) {
+			status = open_lookup(map->type, "",
+					     map->format, map->argc, map->argv, &lookup);
+			if (status != NSS_STATUS_SUCCESS) {
+				map_module_unlock(map);
+				debug(ap->logopt,
+				      "lookup module %s open failed", map->type);
+				return status;
+			}
+			map->lookup = lookup;
 		}
-		map->lookup = lookup;
+		map_module_unlock(map);
 	}
 
+	map_module_readlock(map);
 	lookup = map->lookup;
-
-	master_source_current_wait(ap->entry);
-	ap->entry->current = map;
-
-	status = lookup->lookup_mount(ap, name, name_len, lookup->context);
+	status = lookup->lookup_mount(ap, map, name, name_len, lookup->context);
+	map_module_unlock(map);
 
 	return status;
 }
@@ -930,7 +945,7 @@ static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_
 		return NSS_STATUS_NOTFOUND;
 	}
 
-	if (st.st_mode & __S_IEXEC)
+	if (st.st_mode & S_IEXEC)
 		type = src_prog;
 	else
 		type = src_file;
@@ -1077,7 +1092,7 @@ static struct map_source *lookup_get_map_source(struct master_mapent *entry)
 	if (!S_ISREG(st.st_mode))
 		return NULL;
 
-	if (st.st_mode & __S_IEXEC)
+	if (st.st_mode & S_IEXEC)
 		type = "program";
 	else
 		type = "file";
@@ -1362,11 +1377,11 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 
 			/* If the map hasn't been read (nobrowse
 			 * indirect mounts) then keep cached entries
-			 * for POSITIVE_TIMEOUT.
+			 * for ap->positive_timeout.
 			 */
 			if (!(ap->flags & (MOUNT_FLAG_GHOST |
 					   MOUNT_FLAG_AMD_CACHE_ALL))) {
-				time_t until = me->age + POSITIVE_TIMEOUT;
+				time_t until = me->age + ap->positive_timeout;
 				if ((long) age - (long) until < 0) {
 					me = cache_enumerate(mc, me);
 					continue;
