@@ -32,6 +32,7 @@ struct master *master_list = NULL;
 
 extern const char *global_options;
 extern long global_negative_timeout;
+extern long global_positive_timeout;
 
 /* Attribute to create a joinable thread */
 extern pthread_attr_t th_attr;
@@ -65,6 +66,42 @@ void master_mutex_lock_cleanup(void *arg)
 	return;
 }
 
+void map_module_writelock(struct map_source *map)
+{
+	int status = pthread_rwlock_wrlock(&map->module_lock);
+	if (status)
+		fatal(status);
+}
+
+int map_module_try_writelock(struct map_source *map)
+{
+	int status = pthread_rwlock_trywrlock(&map->module_lock);
+	if (status && status != EBUSY && status != EDEADLK)
+		fatal(status);
+	return status;
+}
+
+void map_module_readlock(struct map_source *map)
+{
+	int status = pthread_rwlock_rdlock(&map->module_lock);
+	if (status)
+		fatal(status);
+}
+
+void map_module_unlock(struct map_source *map)
+{
+	int status = pthread_rwlock_unlock(&map->module_lock);
+	if (status)
+		fatal(status);
+}
+
+void map_module_lock_cleanup(void *arg)
+{
+	struct map_source *map = (struct map_source *) arg;
+
+	map_module_unlock(map);
+}
+
 int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
 			    unsigned nobind, unsigned ghost, int submount)
 {
@@ -76,10 +113,6 @@ int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
 
 	ap->state = ST_INIT;
 
-	ap->state_pipe[0] = -1;
-	ap->state_pipe[1] = -1;
-	ap->logpri_fifo = -1;
-
 	ap->path = strdup(entry->path);
 	if (!ap->path) {
 		free(ap);
@@ -94,11 +127,16 @@ int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
 	/*
 	 * Program command line option overrides config.
 	 * We can't use 0 negative timeout so use default.
+	 * We can't use <0 positive timeout so use default.
 	 */
 	if (global_negative_timeout <= 0)
 		ap->negative_timeout = defaults_get_negative_timeout();
 	else
 		ap->negative_timeout = global_negative_timeout;
+	if (global_positive_timeout < 0)
+		ap->positive_timeout = defaults_get_positive_timeout();
+	else
+		ap->positive_timeout = global_positive_timeout;
 	ap->exp_timeout = defaults_get_timeout();
 	ap->exp_runfreq = 0;
 	ap->flags = 0;
@@ -120,7 +158,6 @@ int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
 
 	ap->parent = NULL;
 	ap->thid = 0;
-	ap->submnt_count = 0;
 	ap->submount = submount;
 	INIT_LIST_HEAD(&ap->mounts);
 	INIT_LIST_HEAD(&ap->submounts);
@@ -155,6 +192,7 @@ master_add_map_source(struct master_mapent *entry,
 	struct map_source *source;
 	char *ntype, *nformat;
 	const char **tmpargv;
+	int status;
 
 	source = malloc(sizeof(struct map_source));
 	if (!source)
@@ -241,6 +279,10 @@ master_add_map_source(struct master_mapent *entry,
 
 	master_source_unlock(entry);
 
+	status = pthread_rwlock_init(&source->module_lock, NULL);
+	if (status)
+		fatal(status);
+
 	return source;
 }
 
@@ -330,6 +372,8 @@ master_get_map_source(struct master_mapent *entry,
 
 static void __master_free_map_source(struct map_source *source, unsigned int free_cache)
 {
+	int status;
+
 	/* instance map sources are not ref counted */
 	if (source->ref && --source->ref)
 		return;
@@ -346,11 +390,14 @@ static void __master_free_map_source(struct map_source *source, unsigned int fre
 
 		instance = source->instance;
 		while (instance) {
-			if (instance->lookup)
+			if (instance->lookup) {
 				close_lookup(instance->lookup);
+				instance->lookup = NULL;
+			}
 			instance = instance->next;
 		}
 		close_lookup(source->lookup);
+		source->lookup = NULL;
 	}
 	if (source->argv)
 		free_argv(source->argc, source->argv);
@@ -363,8 +410,13 @@ static void __master_free_map_source(struct map_source *source, unsigned int fre
 			__master_free_map_source(instance, 0);
 			instance = next;
 		}
+		source->instance = NULL;
 	}
 
+	status = pthread_rwlock_destroy(&source->module_lock);
+	if (status)
+		fatal(status);
+
 	free(source);
 
 	return;
@@ -496,6 +548,10 @@ master_add_source_instance(struct map_source *source, const char *type, const ch
 	if (status)
 		fatal(status);
 
+	status = pthread_rwlock_init(&new->module_lock, NULL);
+	if (status)
+		fatal(status);
+
 	return new;
 }
 
@@ -632,47 +688,6 @@ void master_source_lock_cleanup(void *arg)
 	return;
 }
 
-void master_source_current_wait(struct master_mapent *entry)
-{
-	int status;
-
-	status = pthread_mutex_lock(&entry->current_mutex);
-	if (status) {
-		logmsg("entry current source lock failed");
-		fatal(status);
-	}
-
-	while (entry->current != NULL) {
-		status = pthread_cond_wait(
-				&entry->current_cond, &entry->current_mutex);
-		if (status) {
-			logmsg("entry current source condition wait failed");
-			fatal(status);
-		}
-	}
-
-	return;
-}
-
-void master_source_current_signal(struct master_mapent *entry)
-{
-	int status;
-
-	status = pthread_cond_signal(&entry->current_cond);
-	if (status) {
-		logmsg("entry current source condition signal failed");
-		fatal(status);
-	}
-
-	status = pthread_mutex_unlock(&entry->current_mutex);
-	if (status) {
-		logmsg("entry current source unlock failed");
-		fatal(status);
-	}
-
-	return;
-}
-
 struct master_mapent *master_find_mapent(struct master *master, const char *path)
 {
 	struct list_head *head, *p;
@@ -690,7 +705,35 @@ struct master_mapent *master_find_mapent(struct master *master, const char *path
 	return NULL;
 }
 
-unsigned int master_partial_match_mapent(struct master *master, const char *path)
+struct autofs_point *master_find_mapent_by_devid(struct master *master, dev_t devid)
+{
+	struct autofs_point *ap = NULL;
+	struct list_head *head, *p;
+
+	master_mutex_lock();
+
+	head = &master->mounts;
+	list_for_each(p, head) {
+		struct master_mapent *entry;
+
+		entry = list_entry(p, struct master_mapent, list);
+
+		if (entry->ap->dev == devid) {
+			ap = entry->ap;
+			break;
+		}
+
+		ap = mnt_find_submount_by_devid(&entry->ap->submounts, devid);
+		if (ap)
+			break;
+	}
+
+	master_mutex_unlock();
+
+	return ap;
+}
+
+static unsigned int master_partial_match_amd_mapent(struct master *master, const char *path)
 {
 	struct list_head *head, *p;
 	size_t path_len = strlen(path);
@@ -704,7 +747,7 @@ unsigned int master_partial_match_mapent(struct master *master, const char *path
 
 		entry = list_entry(p, struct master_mapent, list);
 
-		entry_len = strlen(entry->path);
+		entry_len = entry->len;
 		cmp_len = min(entry_len, path_len);
 
 		if (!strncmp(entry->path, path, cmp_len)) {
@@ -755,8 +798,8 @@ struct master_mapent *master_new_mapent(struct master *master, const char *path,
 		return NULL;
 	}
 	entry->path = tmp;
+	entry->len = strlen(tmp);
 
-	entry->thid = 0;
 	entry->age = age;
 	entry->master = master;
 	entry->current = NULL;
@@ -767,14 +810,6 @@ struct master_mapent *master_new_mapent(struct master *master, const char *path,
 	if (status)
 		fatal(status);
 
-	status = pthread_mutex_init(&entry->current_mutex, NULL);
-	if (status)
-		fatal(status);
-
-	status = pthread_cond_init(&entry->current_cond, NULL);
-	if (status)
-		fatal(status);
-
 	INIT_LIST_HEAD(&entry->list);
 
 	return entry;
@@ -789,9 +824,20 @@ void master_add_mapent(struct master *master, struct master_mapent *entry)
 void master_remove_mapent(struct master_mapent *entry)
 {
 	struct master *master = entry->master;
+	struct autofs_point *ap = entry->ap;
 
-	if (entry->ap->submount)
+	if (ap->submount) {
+		struct mnt_list *mnt;
+
+		mnt = mnts_find_submount(ap->path);
+		if (mnt) {
+			warn(ap->logopt,
+			     "map entry %s in use at shutdown", ap->path);
+			mnts_put_mount(mnt);
+		}
+		list_add(&entry->join, &master->completed);
 		return;
+	}
 
 	if (!list_empty(&entry->list)) {
 		list_del_init(&entry->list);
@@ -831,14 +877,6 @@ void master_free_mapent(struct master_mapent *entry)
 	if (status)
 		fatal(status);
 
-	status = pthread_mutex_destroy(&entry->current_mutex);
-	if (status)
-		fatal(status);
-
-	status = pthread_cond_destroy(&entry->current_cond);
-	if (status)
-		fatal(status);
-
 	free(entry);
 
 	return;
@@ -987,7 +1025,7 @@ static void master_add_amd_mount_section_mounts(struct master *master, time_t ag
 		char *map = NULL;
 		char *opts;
 
-		ret = master_partial_match_mapent(master, path);
+		ret = master_partial_match_amd_mapent(master, path);
 		if (ret) {
 			/* If this amd entry is already present in the
 			 * master map it's not a duplicate, don't issue
@@ -1128,12 +1166,14 @@ int master_read_master(struct master *master, time_t age)
 {
 	unsigned int logopt = master->logopt;
 	struct mapent_cache *nc;
+	int ret = 1;
 
 	/*
 	 * We need to clear and re-populate the null map entry cache
 	 * before alowing anyone else to use it.
 	 */
 	wait_for_lookups_and_lock(master);
+	pthread_cleanup_push(master_mutex_lock_cleanup, NULL);
 	if (master->nc) {
 		cache_writelock(master->nc);
 		nc = master->nc;
@@ -1144,7 +1184,8 @@ int master_read_master(struct master *master, time_t age)
 			error(logopt,
 			      "failed to init null map cache for %s",
 			      master->name);
-			return 0;
+			ret = 0;
+			goto done;
 		}
 		cache_writelock(nc);
 		master->nc = nc;
@@ -1160,18 +1201,18 @@ int master_read_master(struct master *master, time_t age)
 		master->read_fail = 0;
 		/* HUP signal sets master->readall == 1 only */
 		if (!master->readall) {
-			master_mutex_unlock();
-			return 0;
+			ret = 0;
+			goto done;
 		} else
 			master_mount_mounts(master, age);
 	}
 
 	if (__master_list_empty(master))
 		warn(logopt, "no mounts in table");
+done:
+	pthread_cleanup_pop(1);
 
-	master_mutex_unlock();
-
-	return 1;
+	return ret;
 }
 
 int master_notify_submount(struct autofs_point *ap, const char *path, enum states state)
@@ -1199,17 +1240,16 @@ int master_notify_submount(struct autofs_point *ap, const char *path, enum state
 		st_wait_task(this->ap, state, 0);
 
 		/*
-		 * If our submount gets to state ST_SHUTDOWN, ST_SHUTDOWN_PENDING or
-		 * ST_SHUTDOWN_FORCE we need to wait until it goes away or changes
-		 * to ST_READY.
+		 * If our submount gets to state ST_SHUTDOWN_PENDING or
+		 * ST_SHUTDOWN_FORCE we need to wait until it goes away
+		 * or changes to state ST_SHUTDOWN or ST_READY.
 		 */
 		st_mutex_lock();
 		while ((sbmnt = mnts_find_submount(path))) {
 			struct timespec t = { 0, 300000000 };
 			struct timespec r;
 
-			if (sbmnt->ap->state != ST_SHUTDOWN &&
-			    sbmnt->ap->state != ST_SHUTDOWN_PENDING &&
+			if (sbmnt->ap->state != ST_SHUTDOWN_PENDING &&
 			    sbmnt->ap->state != ST_SHUTDOWN_FORCE) {
 				ret = 0;
 				mnts_put_mount(sbmnt);
@@ -1311,7 +1351,6 @@ static int master_do_mount(struct master_mapent *entry)
 	}
 
 	suc.ap = ap;
-	suc.root = ap->path;
 	suc.done = 0;
 	suc.status = 0;
 
@@ -1339,7 +1378,7 @@ static int master_do_mount(struct master_mapent *entry)
 		handle_mounts_startup_cond_destroy(&suc);
 		return 0;
 	}
-	entry->thid = thid;
+	ap->thid = thid;
 
 	handle_mounts_startup_cond_destroy(&suc);
 
@@ -1423,9 +1462,6 @@ int master_mount_mounts(struct master *master, time_t age)
 		struct master_mapent *this;
 		struct autofs_point *ap;
 		struct mapent *ne, *nested;
-		struct stat st;
-		int state_pipe, save_errno;
-		int ret;
 
 		this = list_entry(p, struct master_mapent, list);
 		p = p->next;
@@ -1438,7 +1474,7 @@ int master_mount_mounts(struct master *master, time_t age)
 			continue;
 		}
 
-		cache_readlock(nc);
+		cache_writelock(nc);
 		ne = cache_lookup_distinct(nc, this->path);
 		/*
 		 * If this path matched a nulled entry the master map entry
@@ -1447,8 +1483,8 @@ int master_mount_mounts(struct master *master, time_t age)
 		 */
 		if (ne) {
 			int lineno = ne->age;
-			cache_unlock(nc);
 
+			cache_unlock(nc);
 			/* null entry appears after map entry */
 			if (this->maps->master_line < lineno) {
 				warn(ap->logopt,
@@ -1471,30 +1507,19 @@ int master_mount_mounts(struct master *master, time_t age)
 			master_free_mapent(ap->entry);
 			continue;
 		}
+
 		nested = cache_partial_match(nc, this->path);
 		if (nested) {
 			error(ap->logopt,
 			     "removing invalid nested null entry %s",
 			     nested->key);
-			nested = cache_partial_match(nc, this->path);
-			if (nested)
-				cache_delete(nc, nested->key);
+			cache_delete(nc, nested->key);
 		}
 		cache_unlock(nc);
 cont:
-		st_mutex_lock();
-
-		state_pipe = this->ap->state_pipe[1];
-
-		/* No pipe so mount is needed */
-		ret = fstat(state_pipe, &st);
-		save_errno = errno;
-
-		st_mutex_unlock();
-
-		if (!ret)
+		if (ap->thid && ap->state != ST_INIT)
 			check_update_map_sources(this, master->readall);
-		else if (ret == -1 && save_errno == EBADF) {
+		else {
 			if (!master_do_mount(this)) {
 				list_del_init(&this->list);
 				master_free_mapent_sources(ap->entry, 1);
@@ -1908,7 +1933,7 @@ int master_done(struct master *master)
 		entry = list_entry(p, struct master_mapent, join);
 		p = p->next;
 		list_del(&entry->join);
-		pthread_join(entry->thid, NULL);
+		pthread_join(entry->ap->thid, NULL);
 		master_free_mapent_sources(entry, 1);
 		master_free_mapent(entry);
 	}
@@ -1926,6 +1951,9 @@ unsigned int master_get_logopt(void)
 
 int master_kill(struct master *master)
 {
+	if (!master)
+		return 1;
+
 	if (!master_list_empty(master))
 		return 0;
 
@@ -1935,5 +1963,7 @@ int master_kill(struct master *master)
 	cache_release_null_cache(master);
 	free(master);
 
+	master_list = NULL;
+
 	return 1;
 }
