@@ -26,6 +26,7 @@
 #include <sys/wait.h>
 #include <sys/stat.h>
 #include <sys/mount.h>
+#include <pwd.h>
 
 #include "automount.h"
 
@@ -45,7 +46,6 @@ void dump_core(void)
 	sigemptyset(&segv);
 	sigaddset(&segv, SIGSEGV);
 	pthread_sigmask(SIG_UNBLOCK, &segv, NULL);
-	sigprocmask(SIG_UNBLOCK, &segv, NULL);
 
 	raise(SIGSEGV);
 }
@@ -94,7 +94,12 @@ int open_fd(const char *path, int flags)
 #endif
 	fd = open(path, flags);
 	if (fd == -1) {
+		char buf[MAX_ERR_BUF];
+		char *estr;
+
 		open_mutex_unlock();
+		estr = strerror_r(errno, buf, sizeof(buf));
+		logerr("failed to open file: %s", estr);
 		return -1;
 	}
 	check_cloexec(fd);
@@ -113,7 +118,12 @@ int open_fd_mode(const char *path, int flags, int mode)
 #endif
 	fd = open(path, flags, mode);
 	if (fd == -1) {
+		char buf[MAX_ERR_BUF];
+		char *estr;
+
 		open_mutex_unlock();
+		estr = strerror_r(errno, buf, sizeof(buf));
+		logerr("failed to open file: %s", estr);
 		return -1;
 	}
 	check_cloexec(fd);
@@ -123,6 +133,8 @@ int open_fd_mode(const char *path, int flags, int mode)
 
 int open_pipe(int pipefd[2])
 {
+	char buf[MAX_ERR_BUF];
+	char *estr;
 	int ret;
 
 	open_mutex_lock();
@@ -145,6 +157,8 @@ done:
 	return 0;
 err:
 	open_mutex_unlock();
+	estr = strerror_r(errno, buf, sizeof(buf));
+	logerr("failed to open pipe: %s", estr);
 	return -1;
 }
 
@@ -159,7 +173,12 @@ int open_sock(int domain, int type, int protocol)
 #endif
 	fd = socket(domain, type, protocol);
 	if (fd == -1) {
+		char buf[MAX_ERR_BUF];
+		char *estr;
+
 		open_mutex_unlock();
+		estr = strerror_r(errno, buf, sizeof(buf));
+		logerr("failed to open socket: %s", estr);
 		return -1;
 	}
 	check_cloexec(fd);
@@ -184,7 +203,12 @@ FILE *open_fopen_r(const char *path)
 #endif
 	f = fopen(path, "r");
 	if (f == NULL) {
+		char buf[MAX_ERR_BUF];
+		char *estr;
+
 		open_mutex_unlock();
+		estr = strerror_r(errno, buf, sizeof(buf));
+		logerr("failed to open file: %s", estr);
 		return NULL;
 	}
 	check_cloexec(fileno(f));
@@ -209,7 +233,12 @@ FILE *open_setmntent_r(const char *table)
 #endif
 	tab = fopen(table, "r");
 	if (tab == NULL) {
+		char buf[MAX_ERR_BUF];
+		char *estr;
+
 		open_mutex_unlock();
+		estr = strerror_r(errno, buf, sizeof(buf));
+		logerr("failed to open mount table: %s", estr);
 		return NULL;
 	}
 	check_cloexec(fileno(tab));
@@ -306,6 +335,10 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 	struct thread_stdenv_vars *tsv;
 	pid_t euid = 0;
 	gid_t egid = 0;
+	gid_t *groups = NULL;
+	gid_t *saved_groups = NULL;
+	int ngroups = 0;
+	int nsaved_groups = 0;
 
 	if (open_pipe(pipefd))
 		return -1;
@@ -328,6 +361,31 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 	}
 
 	open_mutex_lock();
+
+	if (euid) {
+		struct passwd *pwd;
+
+		pwd = getpwuid(getuid());
+		if (!pwd)
+			fprintf(stderr,
+				"warning: getpwuid: can't get current username\n");
+		else {
+			/* get number of groups for current gid */
+			getgrouplist(pwd->pw_name, getgid(), NULL, &nsaved_groups);
+			saved_groups = malloc(nsaved_groups * sizeof(gid_t));
+
+			/* get current gid groups list */
+			getgrouplist(pwd->pw_name, getgid(), saved_groups, &nsaved_groups);
+		}
+
+		/* get number of groups of mount triggering process */
+		getgrouplist(tsv->user, egid, NULL, &ngroups);
+		groups = malloc(ngroups * sizeof(gid_t));
+
+		/* get groups list of mount triggering process */
+		getgrouplist(tsv->user, egid, groups, &ngroups);
+	}
+
 	f = fork();
 	if (f == 0) {
 		char **pargv = (char **) argv;
@@ -369,10 +427,13 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 				if (!tsv->user)
 					fprintf(stderr,
 						"warning: can't init groups\n");
-				else if (initgroups(tsv->user, egid) == -1)
-					fprintf(stderr,
-						"warning: initgroups: %s\n",
-						strerror(errno));
+				else if (groups) {
+					if (setgroups(ngroups, groups) == -1)
+						fprintf(stderr,
+							"warning: setgroups: %s\n",
+							strerror(errno));
+					free(groups);
+				}
 
 				if (setegid(egid) == -1)
 					fprintf(stderr,
@@ -407,6 +468,11 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 					strerror(errno));
 			if (pgrp >= 0)
 				setpgid(0, pgrp);
+			/* Reset groups for trigger of trailing mount */
+			if (euid && saved_groups) {
+				setgroups(nsaved_groups, saved_groups);
+				free(saved_groups);
+			}
 
 			/*
 			 * The kernel leaves mount type autofs alone because
@@ -445,6 +511,11 @@ done:
 		pthread_sigmask(SIG_SETMASK, &tmpsig, NULL);
 		open_mutex_unlock();
 
+		if (groups)
+			free(groups);
+		if (saved_groups)
+			free(saved_groups);
+
 		close(pipefd[1]);
 
 		if (f < 0) {
@@ -470,7 +541,7 @@ done:
 				while (errp && (p = memchr(sp, '\n', errp))) {
 					*p++ = '\0';
 					if (sp[0])	/* Don't output empty lines */
-						warn(logopt, ">> %s", sp);
+						debug(logopt, ">> %s", sp);
 					errp -= (p - sp);
 					sp = p;
 				}
@@ -481,7 +552,7 @@ done:
 				if (errp >= ERRBUFSIZ) {
 					/* Line too long, split */
 					errbuf[errp] = '\0';
-					warn(logopt, ">> %s", errbuf);
+					debug(logopt, ">> %s", errbuf);
 					errp = 0;
 				}
 			}
@@ -495,7 +566,7 @@ done:
 		if (errp > 0) {
 			/* End of file without \n */
 			errbuf[errp] = '\0';
-			warn(logopt, ">> %s", errbuf);
+			debug(logopt, ">> %s", errbuf);
 		}
 
 		if (waitpid(f, &ret, 0) != f)
