@@ -52,16 +52,6 @@ void st_mutex_unlock(void)
 		fatal(status);
 }
 
-void nextstate(int statefd, enum states next)
-{
-	char buf[MAX_ERR_BUF];
-
-	if (write(statefd, &next, sizeof(next)) != sizeof(next)) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr("write failed %s", estr);
-	}
-}
-
 /*
  * Handle expire thread cleanup and return the next state the system
  * should enter as a result.
@@ -99,36 +89,14 @@ void expire_cleanup(void *arg)
 			/*
 			 * If we're a submount and we've just pruned or
 			 * expired everything away, try to shut down.
-			 *
-			 * Since we use the the fact that a mount will not
-			 * expire for at least ap->exp_timeout to avoid a
-			 * mount <-> expire race we need to wait before
-			 * letting a submount expire away. We also need
-			 * them to go away fairly quickly so the owner
-			 * mount expires in a reasonable time. Just skip
-			 * one expire check after it's no longer busy before
-			 * allowing it to shutdown.
-			 *
-			 * But if this mount point is an amd format map it
-			 * is better to keep the mount around longer. This
-			 * is because of the common heavy reuse of maps in
-			 * amd maps and we want to try and avoid constantly
-			 * re-reading large maps.
 			 */
 			if (ap->submount && !success) {
 				rv = ops->askumount(ap->logopt, ap->ioctlfd, &idle);
-				if (!rv && idle && ap->submount > 1) {
-					struct map_source *map = ap->entry->maps;
-
-					if (ap->submount > 4 ||
-					   !(map->flags & MAP_FLAG_FORMAT_AMD)) {
-						next = ST_SHUTDOWN_PENDING;
-						break;
-					}
+				if (!rv && idle) {
+					next = ST_SHUTDOWN_PENDING;
+					break;
 				}
-				ap->submount++;
-			} else if (ap->submount > 1)
-				ap->submount = 1;
+			}
 
 			if (ap->state == ST_EXPIRE)
 				conditional_alarm_add(ap, ap->exp_runfreq);
@@ -333,16 +301,20 @@ static int do_readmap_mount(struct autofs_point *ap,
 
 	nc = ap->entry->master->nc;
 
+	cache_writelock(nc);
 	ne = cache_lookup_distinct(nc, me->key);
-	if (!ne) {
+	if (ne)
+		cache_unlock(nc);
+	else {
 		nested = cache_partial_match(nc, me->key);
-		if (nested) {
+		if (!nested)
+			cache_unlock(nc);
+		else {
+			cache_delete(nc, nested->key);
+			cache_unlock(nc);
 			error(ap->logopt,
 			      "removing invalid nested null entry %s",
 			      nested->key);
-			nested = cache_partial_match(nc, me->key);
-			if (nested)
-				cache_delete(nc, nested->key);
 		}
 	}
 
@@ -394,7 +366,7 @@ static int do_readmap_mount(struct autofs_point *ap,
 			valid->ioctlfd = me->ioctlfd;
 			me->ioctlfd = -1;
 			/* Set device and inode number of the new mapent */
-			cache_set_ino_index(vmc, me->key, me->dev, me->ino);
+			cache_set_ino_index(vmc, me);
 			cache_unlock(vmc);
 			/* Set timeout and calculate the expire run frequency */
 			timeout = get_exp_timeout(ap, map);
@@ -421,7 +393,7 @@ static void *do_readmap(void *arg)
 {
 	struct autofs_point *ap;
 	struct map_source *map;
-	struct mapent_cache *nc, *mc;
+	struct mapent_cache *mc;
 	struct readmap_args *ra;
 	int status;
 	time_t now;
@@ -461,14 +433,11 @@ static void *do_readmap(void *arg)
 		ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
 		ops->timeout(ap->logopt, ap->ioctlfd, timeout);
 		lookup_prune_cache(ap, now);
-		status = lookup_ghost(ap, ap->path);
+		status = lookup_ghost(ap);
 	} else {
 		struct mapent *me;
 		unsigned int append_alarm = !ap->exp_runfreq;
 
-		nc = ap->entry->master->nc;
-		cache_readlock(nc);
-		pthread_cleanup_push(cache_lock_cleanup, nc);
 		master_source_readlock(ap->entry);
 		pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 		map = ap->entry->maps;
@@ -506,7 +475,6 @@ restart:
 		}
 
 		pthread_cleanup_pop(1);
-		pthread_cleanup_pop(1);
 	}
 
 	pthread_cleanup_pop(1);
@@ -653,12 +621,16 @@ static unsigned int st_force_shutdown(struct autofs_point *ap)
 
 static unsigned int st_shutdown(struct autofs_point *ap)
 {
+	int ret;
+
 	debug(ap->logopt, "state %d path %s", ap->state, ap->path);
 
 	assert(ap->state == ST_SHUTDOWN_PENDING || ap->state == ST_SHUTDOWN_FORCE);
 
 	ap->state = ST_SHUTDOWN;
-	nextstate(ap->state_pipe[1], ST_SHUTDOWN);
+	ret = pthread_kill(ap->thid, SIGCONT);
+	if (ret)
+		error(LOGOPT_ANY, "error %d sending shutdown signal", ret);
 
 	return 0;
 }
