@@ -17,6 +17,8 @@
 #ifndef LOG_H
 #define LOG_H
 
+#include <unistd.h>
+
 /* Define logging functions */
 
 #define LOGOPT_NONE	0x0000
@@ -29,12 +31,16 @@ struct autofs_point;
 
 extern void set_log_norm(void);
 extern void set_log_verbose(void);
-extern void set_log_debug(void);
+extern void set_log_debug(int level);
 extern void set_log_norm_ap(struct autofs_point *ap);
 extern void set_log_verbose_ap(struct autofs_point *ap);
 extern void set_log_debug_ap(struct autofs_point *ap);
 extern void set_mnt_logging(unsigned global_logopt);
 
+extern unsigned int have_log_verbose(void);
+extern unsigned int have_log_debug(void);
+extern int get_log_debug_level(void);
+
 extern void open_log(void);
 extern void log_to_stderr(void);
  
