@@ -65,8 +65,6 @@ struct amd_entry {
 	char *umount;
 	struct selector *selector;
 	struct list_head list;
-	struct list_head entries;
-	struct list_head ext_mount;
 };
 
 int amd_parse_list(struct autofs_point *,
