@@ -47,9 +47,11 @@
 
 #define NAME_MASTER_MAP			"master_map_name"
 
+#define NAME_OPEN_FILE_LIMIT		"open_file_limit"
 #define NAME_TIMEOUT			"timeout"
 #define NAME_MASTER_WAIT		"master_wait"
 #define NAME_NEGATIVE_TIMEOUT		"negative_timeout"
+#define NAME_POSITIVE_TIMEOUT		"positive_timeout"
 #define NAME_BROWSE_MODE		"browse_mode"
 #define NAME_LOGGING			"logging"
 #define NAME_FORCE_STD_PROG_MAP_ENV	"force_standard_program_map_env"
@@ -290,6 +292,11 @@ static int conf_load_autofs_defaults(void)
 	const char *sec = autofs_gbl_sec;
 	int ret;
 
+	ret = conf_update(sec, NAME_OPEN_FILE_LIMIT,
+			  DEFAULT_OPEN_FILE_LIMIT, CONF_ENV);
+	if (ret == CFG_FAIL)
+		goto error;
+
 	ret = conf_update(sec, NAME_TIMEOUT,
 			  DEFAULT_TIMEOUT, CONF_ENV);
 	if (ret == CFG_FAIL)
@@ -1670,6 +1677,17 @@ int defaults_master_set(void)
 	return 0;
 }
 
+unsigned long defaults_get_open_file_limit(void)
+{
+	long limit;
+
+	limit = conf_get_number(autofs_gbl_sec, NAME_OPEN_FILE_LIMIT);
+	if (limit < 0)
+		limit = atol(DEFAULT_OPEN_FILE_LIMIT);
+
+	return limit;
+}
+
 unsigned int defaults_get_timeout(void)
 {
 	long timeout;
@@ -1703,6 +1721,17 @@ unsigned int defaults_get_negative_timeout(void)
 	return (unsigned int) n_timeout;
 }
 
+unsigned int defaults_get_positive_timeout(void)
+{
+	long p_timeout;
+
+	p_timeout = conf_get_number(autofs_gbl_sec, NAME_POSITIVE_TIMEOUT);
+	if (p_timeout <= 0)
+		p_timeout = atol(DEFAULT_POSITIVE_TIMEOUT);
+
+	return (unsigned int) p_timeout;
+}
+
 unsigned int defaults_get_browse_mode(void)
 {
 	int res;
