@@ -231,6 +231,32 @@ unsigned int get_kver_minor(void)
 	return kver.minor;
 }
 
+int open_ioctlfd(struct autofs_point *ap, const char *path, dev_t dev)
+{
+	struct ioctl_ops *ops = get_ioctl_ops();
+	int fd = -1;
+	int error;
+
+	error = ops->open(ap->logopt, &fd, dev, path);
+	if (error == -1) {
+		char buf[MAX_ERR_BUF];
+		int err = errno;
+		char *estr;
+
+		if (errno == ENOENT)
+			return -1;
+
+		estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		error(ap->logopt,
+		      "failed to open ioctlfd for %s, error: %s",
+		      path, estr);
+		errno = err;
+		return -1;
+	}
+
+	return fd;
+}
+
 #ifdef HAVE_MOUNT_NFS
 static int extract_version(char *start, struct nfs_mount_vers *vers)
 {
@@ -1059,6 +1085,40 @@ struct mnt_list *mnts_find_submount(const char *path)
 	return NULL;
 }
 
+static struct autofs_point *__mnt_find_submount_by_devid(struct list_head *submounts, dev_t devid)
+{
+	struct autofs_point *ap = NULL;
+	struct list_head *p;
+
+	list_for_each(p, submounts) {
+		struct mnt_list *this;
+
+		this = list_entry(p, struct mnt_list, submount);
+
+		if (this->ap->dev == devid) {
+			ap = this->ap;
+			break;
+		}
+
+		ap = mnt_find_submount_by_devid(&this->ap->submounts, devid);
+		if (ap)
+			break;
+	}
+
+	return ap;
+}
+
+struct autofs_point *mnt_find_submount_by_devid(struct list_head *submounts, dev_t devid)
+{
+	struct autofs_point *ap = NULL;
+
+	mnts_hash_mutex_lock();
+	ap = __mnt_find_submount_by_devid(submounts, devid);
+	mnts_hash_mutex_unlock();
+
+	return ap;
+}
+
 struct mnt_list *mnts_add_submount(struct autofs_point *ap)
 {
 	struct mnt_list *this;
@@ -1341,7 +1401,7 @@ static struct tree_node *tree_add_node(struct tree_node *root, void *ptr)
 	}
 
 	if (!eq)
-		error(LOGOPT_ANY, "cannot add duplicate entry to tree");
+		return p;
 	else {
 		if (eq < 0)
 			return tree_add_left(p, ptr);
@@ -1515,8 +1575,10 @@ static int tree_host_cmp(struct tree_node *n, void *ptr)
 	int eq;
 
 	eq = strcmp(exp->dir, n_exp->dir);
-	if (!eq)
+	if (!eq) {
+		error(LOGOPT_ANY, "duplicate entry %s ignored", exp->dir);
 		return 0;
+	}
 	return (exp_len < n_exp_len) ? -1 : 1;
 }
 
@@ -1664,16 +1726,26 @@ static int tree_mapent_delete_offset_tree(struct tree_node *root)
 	 */
 	if (MAPENT_ROOT(me) != MAPENT_NODE(me)) {
 		struct tree_node *root = MAPENT_ROOT(me);
+		char *key;
 
-		debug(logopt, "deleting offset key %s", me->key);
+		key = strdup(me->key);
+		if (!key) {
+			char buf[MAX_ERR_BUF];
+			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			error(logopt, "strdup: %s", estr);
+			return 0;
+		}
+
+		debug(logopt, "deleting offset key %s", key);
 
 		/* cache_delete won't delete an active offset */
 		MAPENT_SET_ROOT(me, NULL);
-		ret = cache_delete(me->mc, me->key);
+		ret = cache_delete(me->mc, key);
 		if (ret != CHE_OK) {
 			MAPENT_SET_ROOT(me, root);
-			warn(logopt, "failed to delete offset %s", me->key);
+			warn(logopt, "failed to delete offset %s", key);
 		}
+		free(key);
 	} else {
 		MAPENT_SET_ROOT(me, NULL);
 		MAPENT_SET_PARENT(me, NULL);
@@ -1831,7 +1903,7 @@ static int tree_mapent_umount_offset(struct mapent *oe, void *ptr)
 	 * Check for and umount subtree offsets resulting from
 	 * nonstrict mount fail.
 	 */
-	ret = tree_mapent_umount_offsets(oe, ctxt->strict);
+	ret = tree_mapent_umount_offsets(oe);
 	if (!ret)
 		return 0;
 
@@ -1857,8 +1929,7 @@ static int tree_mapent_umount_offset(struct mapent *oe, void *ptr)
 	 */
 	if (oe->ioctlfd != -1 ||
 	    is_mounted(oe->key, MNTS_REAL)) {
-		if (umount_ent(ap, oe->key) &&
-		    is_mounted(oe->key, MNTS_REAL)) {
+		if (umount_ent(ap, oe->key)) {
 			debug(ap->logopt,
 			      "offset %s has active mount, invalidate",
 			      oe->key);
@@ -1940,7 +2011,7 @@ static int tree_mapent_mount_offsets_work(struct tree_node *n, void *ptr)
 			tree_mapent_mount_offsets(oe, !ctxt->strict);
 	}
 
-	return ret;
+	return (ctxt->strict ? ret : 1);
 }
 
 int tree_mapent_mount_offsets(struct mapent *oe, int nonstrict)
@@ -1963,14 +2034,14 @@ static int tree_mapent_umount_offsets_work(struct tree_node *n, void *ptr)
 	return tree_mapent_umount_offset(oe, ptr);
 }
 
-int tree_mapent_umount_offsets(struct mapent *oe, int nonstrict)
+int tree_mapent_umount_offsets(struct mapent *oe)
 {
 	struct tree_node *base = MAPENT_NODE(oe);
 	struct autofs_point *ap = oe->mc->ap;
 	struct traverse_subtree_context ctxt = {
 		.ap = ap,
 		.base = base,
-		.strict = !nonstrict,
+		.strict = 1,
 	};
 	int ret;
 
@@ -1998,8 +2069,7 @@ int tree_mapent_umount_offsets(struct mapent *oe, int nonstrict)
 		 */
 		if (is_mounted(mp, MNTS_REAL)) {
 			info(ap->logopt, "unmounting dir = %s", mp);
-			if (umount_ent(ap, mp) &&
-			    is_mounted(mp, MNTS_REAL)) {
+			if (umount_ent(ap, mp)) {
 				if (!tree_mapent_mount_offsets(oe, 1))
 					warn(ap->logopt,
 					     "failed to remount offset triggers");
@@ -2169,11 +2239,8 @@ struct mnt_list *get_mnt_list(const char *path, int include)
 		return NULL;
 
 	tab = open_fopen_r(_PROC_MOUNTS);
-	if (!tab) {
-		char *estr = strerror_r(errno, buf, PATH_MAX - 1);
-		logerr("fopen: %s", estr);
+	if (!tab)
 		return NULL;
-	}
 
 	while ((mnt = local_getmntent_r(tab, &mnt_wrk, buf, PATH_MAX * 3))) {
 		len = strlen(mnt->mnt_dir);
@@ -2280,11 +2347,8 @@ static int table_is_mounted(const char *mp, unsigned int type)
 		return 0;
 
 	tab = open_fopen_r(_PROC_MOUNTS);
-	if (!tab) {
-		char *estr = strerror_r(errno, buf, PATH_MAX - 1);
-		logerr("fopen: %s", estr);
+	if (!tab)
 		return 0;
-	}
 
 	while ((mnt = local_getmntent_r(tab, &mnt_wrk, buf, PATH_MAX * 3))) {
 		size_t len = strlen(mnt->mnt_dir);
@@ -2379,11 +2443,17 @@ void set_tsd_user_vars(unsigned int logopt, uid_t uid, gid_t gid)
 
 	/* Try to get passwd info */
 
+	/* sysconf may return -1 with unchanged errno to indicate unlimited
+	 * size, same for the call for _SC_GETGR_R_SIZE_MAX  below
+	 */
+	errno = 0;
 	tmplen = sysconf(_SC_GETPW_R_SIZE_MAX);
-	if (tmplen < 0) {
+	if (tmplen < 0 && errno != 0) {
 		error(logopt, "failed to get buffer size for getpwuid_r");
 		goto free_tsv;
 	}
+	if (tmplen < 0)
+		tmplen = 1024;	/* assume something reasonable */
 
 	pw_tmp = malloc(tmplen + 1);
 	if (!pw_tmp) {
@@ -2416,11 +2486,14 @@ void set_tsd_user_vars(unsigned int logopt, uid_t uid, gid_t gid)
 
 	/* Try to get group info */
 
+	errno = 0;
 	grplen = sysconf(_SC_GETGR_R_SIZE_MAX);
-	if (grplen < 0) {
+	if (grplen < 0 && errno != 0) {
 		error(logopt, "failed to get buffer size for getgrgid_r");
 		goto free_tsv_home;
 	}
+	if (grplen < 0)
+		grplen = 1024;
 
 	gr_tmp = NULL;
 	status = ERANGE;
@@ -2672,7 +2745,7 @@ static int remount_active_mount(struct autofs_point *ap,
 	*ioctlfd = -1;
 
 	/* Open failed, no mount present */
-	ops->open(ap->logopt, &fd, devid, path);
+	fd = open_ioctlfd(ap, path, devid);
 	if (fd == -1)
 		return REMOUNT_OPEN_FAIL;
 
@@ -2714,10 +2787,20 @@ static int remount_active_mount(struct autofs_point *ap,
 		ops->close(ap->logopt, fd);
 		return REMOUNT_STAT_FAIL;
 	}
-	if (type != t_indirect)
-		cache_set_ino_index(me->mc, path, st.st_dev, st.st_ino);
-	else
+	if (type == t_indirect)
 		ap->dev = st.st_dev;
+	else {
+		if (strcmp(path, me->key)) {
+			error(ap->logopt,
+			      "mount point path mismatch, path %s mapent %s", path, me->key);
+			debug(ap->logopt, "couldn't re-connect to mount %s", path);
+			ops->close(ap->logopt, fd);
+			return REMOUNT_STAT_FAIL;
+		}
+		me->dev = st.st_dev;
+		me->ino = st.st_ino;
+		cache_set_ino_index(me->mc, me);
+	}
 	notify_mount_result(ap, path, timeout, str_type);
 
 	*ioctlfd = fd;
@@ -2760,6 +2843,7 @@ static int remount_active_mount(struct autofs_point *ap,
 int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
+	struct mapent *mapent;
 	const char *path;
 	int ret, fd;
 	dev_t devid;
@@ -2790,14 +2874,31 @@ int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 			ap->flags &= ~MOUNT_FLAG_DIR_CREATED;
 		else
 			ap->flags |= MOUNT_FLAG_DIR_CREATED;
-	} else {
-		me->flags &= ~MOUNT_FLAG_DIR_CREATED;
-		if (type == t_offset) {
-			if (!is_mounted(MM_PARENT(me)->key, MNTS_REAL))
-				me->flags |= MOUNT_FLAG_DIR_CREATED;
-		}
+		goto done;
 	}
 
+	me->flags &= ~MOUNT_FLAG_DIR_CREATED;
+	mapent = IS_MM(me) ? MM_PARENT(me) : me;
+	/* Direct or offset mount, key is full path */
+	if (mapent->key[0] == '/') {
+		if (!is_mounted(mapent->key, MNTS_REAL))
+			me->flags |= MOUNT_FLAG_DIR_CREATED;
+	} else {
+		char *p_key = mapent->key;
+		char mp[PATH_MAX + 1];
+		int len;
+
+		len = mount_fullpath(mp, PATH_MAX, ap->path, ap->len, p_key);
+		if (len > PATH_MAX) {
+			/* This should never happen due to earlier checks */
+			error(ap->logopt, "mountpoint path too long");
+			return 0;
+		}
+
+		if (!is_mounted(mp, MNTS_REAL))
+			me->flags |= MOUNT_FLAG_DIR_CREATED;
+	}
+done:
 	/*
 	 * Either we opened the mount or we're re-reading the map.
 	 * If we opened the mount and ioctlfd is not -1 we have
@@ -2843,10 +2944,9 @@ static int set_mount_catatonic(struct autofs_point *ap, struct mapent *me, int i
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	unsigned int opened = 0;
-	char buf[MAX_ERR_BUF];
-	char *path;
-	int fd = -1;
-	int error;
+	const char *path;
+	int fd;
+	int err;
 	dev_t dev;
 
 	path = ap->path;
@@ -2861,44 +2961,31 @@ static int set_mount_catatonic(struct autofs_point *ap, struct mapent *me, int i
 	else if (me && me->ioctlfd >= 0)
 		fd = me->ioctlfd;
 	else {
-		error = ops->open(ap->logopt, &fd, dev, path);
-		if (error == -1) {
-			int err = errno;
-			char *estr;
-
-			if (errno == ENOENT)
-				return 0;
-
-			estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			error(ap->logopt,
-			      "failed to open ioctlfd for %s, error: %s",
-			      path, estr);
-			return err;
-		}
+		fd = open_ioctlfd(ap, path, dev);
+		if (fd == -1)
+			return (errno == ENOENT ? 0 : errno);
 		opened = 1;
 	}
 
-	if (fd >= 0) {
-		error = ops->catatonic(ap->logopt, fd);
-		if (error == -1) {
-			int err = errno;
-			char *estr;
+	err = ops->catatonic(ap->logopt, fd);
+	if (err == -1) {
+		char buf[MAX_ERR_BUF];
+		char *estr;
 
-			estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			error(ap->logopt,
-			      "failed to set %s catatonic, error: %s",
-			      path, estr);
-			if (opened)
-				ops->close(ap->logopt, fd);
-			return err;
-		}
-		if (opened)
-			ops->close(ap->logopt, fd);
+		err = errno;
+		estr = strerror_r(err, buf, MAX_ERR_BUF);
+		error(ap->logopt,
+		      "failed to set %s catatonic, error: %s",
+		      path, estr);
+		goto out;
 	}
 
 	debug(ap->logopt, "set %s catatonic", path);
+out:
+	if (opened)
+		ops->close(ap->logopt, fd);
 
-	return 0;
+	return err;
 }
 
 static int set_offset_tree_catatonic_work(struct tree_node *n, void *ptr)
@@ -2967,6 +3054,7 @@ void set_direct_mount_tree_catatonic(struct autofs_point *ap, struct mapent *me)
 
 int umount_ent(struct autofs_point *ap, const char *path)
 {
+	unsigned int mounted;
 	int rv;
 
 	if (ap->state != ST_SHUTDOWN_FORCE)
@@ -2978,6 +3066,8 @@ int umount_ent(struct autofs_point *ap, const char *path)
 		rv = spawn_umount(ap->logopt, "-l", path, NULL);
 	}
 
+	mounted = is_mounted(path, MNTS_REAL);
+
 	if (rv && (ap->state == ST_SHUTDOWN_FORCE || ap->state == ST_SHUTDOWN)) {
 		/*
 		 * Verify that we actually unmounted the thing.  This is a
@@ -2989,20 +3079,20 @@ int umount_ent(struct autofs_point *ap, const char *path)
 		 * so that we do not try to call rmdir_path on the
 		 * directory.
 		 */
-		if (is_mounted(path, MNTS_REAL)) {
+		if (mounted) {
 			crit(ap->logopt,
 			     "the umount binary reported that %s was "
 			     "unmounted, but there is still something "
 			     "mounted on this path.", path);
-			rv = -1;
+			mounted = -1;
 		}
 	}
 
-	/* On success, check for mounted mount and remove it if found */
-	if (!rv)
+	/* If mount is gone remove it from mounted mounts list. */
+	if (!mounted)
 		mnts_remove_mount(path, MNTS_MOUNTED);
 
-	return rv;
+	return mounted;
 }
 
 int umount_amd_ext_mount(struct autofs_point *ap, const char *path)
