@@ -218,7 +218,7 @@ unsigned int get_proximity(struct sockaddr *host_addr)
 	int addr_len;
 	char buf[MAX_ERR_BUF];
 	uint32_t mask, ha, ia, *mask6, *ha6, *ia6;
-	int ret;
+	int ret, at_least_one;
 
 	addr = NULL;
 	addr6 = NULL;
@@ -228,6 +228,7 @@ unsigned int get_proximity(struct sockaddr *host_addr)
 	ha6 = NULL;
 	ia6 = NULL;
 	ha = 0;
+	at_least_one = 0;
 
 	switch (host_addr->sa_family) {
 	case AF_INET:
@@ -245,6 +246,14 @@ unsigned int get_proximity(struct sockaddr *host_addr)
 		hst6_addr = (struct in6_addr *) &addr6->sin6_addr;
 		ha6 = &hst6_addr->s6_addr32[0];
 		addr_len = sizeof(*hst6_addr);
+
+		/* The link-local address always seems to be a problem so
+		 * ignore it when trying to work out if the address we have
+		 * is reachable.
+		 */
+		if (IN6_IS_ADDR_LINKLOCAL(hst6_addr))
+			return PROXIMITY_UNSUPPORTED;
+
 		break;
 #endif
 
@@ -278,6 +287,14 @@ unsigned int get_proximity(struct sockaddr *host_addr)
 				freeifaddrs(ifa);
 				return PROXIMITY_LOCAL;
 			}
+
+			/* If the target address is the loopback address it will
+			 * have matched above so we can ignore it when trying to
+			 * work out if the address we have is reachable.
+			 */
+			if (addr->sin_addr.s_addr != INADDR_LOOPBACK)
+				at_least_one = 1;
+
 			break;
 
 		case AF_INET6:
@@ -290,6 +307,15 @@ unsigned int get_proximity(struct sockaddr *host_addr)
 				freeifaddrs(ifa);
 				return PROXIMITY_LOCAL;
 			}
+
+			/* If the interface address is the loopback address it will
+			 * have matched above so we can ignore it and the link-local
+			 * address always seems to be a problem so ignore it too when
+			 * trying to work out if the address we have is reachable.
+			 */
+			if (!IN6_IS_ADDR_LINKLOCAL(&if6_addr->sin6_addr) &&
+			    !IN6_IS_ADDR_LOOPBACK(&if6_addr->sin6_addr))
+				at_least_one = 1;
 #endif
 		default:
 			break;
@@ -297,6 +323,11 @@ unsigned int get_proximity(struct sockaddr *host_addr)
 		this = this->ifa_next;
 	}
 
+	if (!at_least_one) {
+		freeifaddrs(ifa);
+		return PROXIMITY_UNSUPPORTED;
+	}
+
 	this = ifa;
 	while (this) {
 		if (!(this->ifa_flags & IFF_UP) ||
@@ -353,6 +384,9 @@ unsigned int get_proximity(struct sockaddr *host_addr)
 			if6_addr = (struct sockaddr_in6 *) this->ifa_addr;
 			ia6 = &if6_addr->sin6_addr.s6_addr32[0];
 
+			if (IN6_IS_ADDR_LINKLOCAL(&if6_addr->sin6_addr))
+				break;
+
 			/* Is the address within the network of the interface */
 
 			msk6_addr = (struct sockaddr_in6 *) this->ifa_netmask;
@@ -532,8 +566,11 @@ struct mapent *match_cached_key(struct autofs_point *ap,
 			while ((me = cache_lookup_key_next(me)))
 				if (me->source == source)
 					break;
-			if (!me)
+			if (!me) {
 				me = cache_lookup_distinct(mc, "*");
+				if (me && (me->source != source))
+					me = NULL;
+			}
 		}
 
 		if (!me)
@@ -545,7 +582,9 @@ struct mapent *match_cached_key(struct autofs_point *ap,
 		 */
 		if (!(ap->flags & MOUNT_FLAG_REMOUNT) &&
 		    ap->type == LKP_INDIRECT && *me->key == '*') {
-			ret = cache_update(mc, source, key, me->mapent, me->age);
+			time_t now = monotonic_time(NULL);
+
+			ret = cache_update(mc, source, key, me->mapent, now);
 			if (!(ret & (CHE_OK | CHE_UPDATED)))
 				me = NULL;
 		}
@@ -873,6 +912,19 @@ char *sanitize_path(const char *path, int origlen, unsigned int type, unsigned i
 	return s_path;
 }
 
+int starts_with_null_opt(const char *str)
+{
+	if (str && strlen(str) >= 5 && *str == '-') {
+		char sep = *(str + 5);
+
+		if (sep == 0 || sep == ' ' || sep == ',') {
+			if (!strncmp(str, "-null", 5))
+				return 1;
+		}
+	}
+	return 0;
+}
+
 static char *hasopt(const char *str, const char *opt)
 {
 	const size_t optlen = strlen(opt);
