@@ -338,10 +338,13 @@ static int setautomntent_wait(unsigned int logopt,
 	     "can't connect to sssd, retry for %d seconds",
 	     retries);
 
-	while (++retry <= retries) {
+	while (++retry < retries) {
 		struct timespec t = { SSS_WAIT_INTERVAL, 0 };
 		struct timespec r;
 
+		while (nanosleep(&t, &r) == -1 && errno == EINTR)
+			memcpy(&t, &r, sizeof(struct timespec));
+
 		ret = ctxt->setautomntent(ctxt->mapname, sss_ctxt);
 		if (proto_version(ctxt) == 0) {
 			if (ret != ENOENT)
@@ -355,9 +358,6 @@ static int setautomntent_wait(unsigned int logopt,
 			free(*sss_ctxt);
 			*sss_ctxt = NULL;
 		}
-
-		while (nanosleep(&t, &r) == -1 && errno == EINTR)
-			memcpy(&t, &r, sizeof(struct timespec));
 	}
 
 	if (!ret)
@@ -368,7 +368,7 @@ static int setautomntent_wait(unsigned int logopt,
 			*sss_ctxt = NULL;
 		}
 
-		if (proto_version(ctxt) == 0 && retry > retries)
+		if (proto_version(ctxt) == 0 && retry >= retries)
 			ret = ETIMEDOUT;
 	}
 	return ret;
@@ -394,7 +394,17 @@ static int setautomntent(unsigned int logopt,
 			if (ret != ENOENT)
 				goto error;
 		} else {
-			if (ret != ENOENT && ret != EHOSTDOWN)
+			/* If we get an ENOENT here assume it's accurrate
+			 * and return the error.
+			 */
+			if (ret == ENOENT) {
+				error(logopt, MODPREFIX
+				      "setautomountent: entry for map %s not found",
+				      ctxt->mapname);
+				err = NSS_STATUS_NOTFOUND;
+				goto free;
+			}
+			if (ret != EHOSTDOWN)
 				goto error;
 		}
 
@@ -410,6 +420,10 @@ static int setautomntent(unsigned int logopt,
 			if (ret == EINVAL)
 				goto free;
 			if (ret == ENOENT) {
+				/* Map info. not found after host became available */
+				error(logopt, MODPREFIX
+				      "setautomountent: entry for map %s not found",
+				      ctxt->mapname);
 				err = NSS_STATUS_NOTFOUND;
 				goto free;
 			}
@@ -461,10 +475,13 @@ static int getautomntent_wait(unsigned int logopt,
 	 "can't contact sssd to to get map entry, retry for %d seconds",
 	 retries);
 
-	while (++retry <= retries) {
+	while (++retry < retries) {
 		struct timespec t = { SSS_WAIT_INTERVAL, 0 };
 		struct timespec r;
 
+		while (nanosleep(&t, &r) == -1 && errno == EINTR)
+			memcpy(&t, &r, sizeof(struct timespec));
+
 		ret = ctxt->getautomntent_r(key, value, sss_ctxt);
 		if (proto_version(ctxt) == 0) {
 			if (ret != ENOENT)
@@ -473,16 +490,13 @@ static int getautomntent_wait(unsigned int logopt,
 			if (ret != EHOSTDOWN)
 				break;
 		}
-
-		while (nanosleep(&t, &r) == -1 && errno == EINTR)
-			memcpy(&t, &r, sizeof(struct timespec));
 	}
 
 	if (!ret)
 		info(logopt,
 		     "successfully contacted sssd to get map entry");
 	else {
-		if (retry == retries)
+		if (proto_version(ctxt) == 0 && retry >= retries)
 			ret = ETIMEDOUT;
 	}
 	return ret;
@@ -586,10 +600,13 @@ static int getautomntbyname_wait(unsigned int logopt,
 	"can't contact sssd to to lookup key value, retry for %d seconds",
 	retries);
 
-	while (++retry <= retries) {
+	while (++retry < retries) {
 		struct timespec t = { SSS_WAIT_INTERVAL, 0 };
 		struct timespec r;
 
+		while (nanosleep(&t, &r) == -1 && errno == EINTR)
+			memcpy(&t, &r, sizeof(struct timespec));
+
 		ret = ctxt->getautomntbyname_r(key, value, sss_ctxt);
 		if (proto_version(ctxt) == 0) {
 			if (ret != ENOENT)
@@ -598,16 +615,13 @@ static int getautomntbyname_wait(unsigned int logopt,
 			if (ret != EHOSTDOWN)
 				break;
 		}
-
-		while (nanosleep(&t, &r) == -1 && errno == EINTR)
-			memcpy(&t, &r, sizeof(struct timespec));
 	}
 
 	if (!ret)
 		info(logopt,
 		     "successfully contacted sssd to lookup key value");
 	else {
-		if (proto_version(ctxt) == 0 && retry > retries)
+		if (proto_version(ctxt) == 0 && retry >= retries)
 			ret = ETIMEDOUT;
 	}
 	return ret;
@@ -757,23 +771,17 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	return NSS_STATUS_SUCCESS;
 }
 
-int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
+int lookup_read_map(struct autofs_point *ap, struct map_source *map, time_t age, void *context)
 {
 	struct lookup_context *ctxt = (struct lookup_context *) context;
-	struct map_source *source;
-	struct mapent_cache *mc;
+	struct map_source *source = map;
+	struct mapent_cache *mc = source->mc;
 	void *sss_ctxt = NULL;
 	char *key;
 	char *value = NULL;
 	char *s_key;
 	int count, ret;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
-	mc = source->mc;
-
 	/*
 	 * If we don't need to create directories (or don't need
 	 * to read an amd cache:=all map) then there's no use
@@ -859,11 +867,10 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 	return NSS_STATUS_SUCCESS;
 }
 
-static int lookup_one(struct autofs_point *ap,
-		char *qKey, int qKey_len, struct lookup_context *ctxt)
+static int lookup_one(struct autofs_point *ap, struct map_source *source,
+		      char *qKey, int qKey_len, struct lookup_context *ctxt)
 {
-	struct map_source *source;
-	struct mapent_cache *mc;
+	struct mapent_cache *mc = source->mc;
 	struct mapent *we;
 	void *sss_ctxt = NULL;
 	time_t age = monotonic_time(NULL);
@@ -871,12 +878,6 @@ static int lookup_one(struct autofs_point *ap,
 	char *s_key;
 	int err, ret;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
-	mc = source->mc;
-
 	ret = setautomntent(ap->logopt, ctxt, &sss_ctxt, SSS_LOOKUP_KEY);
 	if (ret)
 		return ret;
@@ -969,24 +970,14 @@ wild:
 }
 
 static int check_map_indirect(struct autofs_point *ap,
-			      char *key, int key_len,
+			      struct map_source *source, char *key, int key_len,
 			      struct lookup_context *ctxt)
 {
-	struct map_source *source;
-	struct mapent_cache *mc;
+	struct mapent_cache *mc = source->mc;
 	int ret, cur_state;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
-	mc = source->mc;
-
-	master_source_current_wait(ap->entry);
-	ap->entry->current = source;
-
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-	ret = lookup_one(ap, key, key_len, ctxt);
+	ret = lookup_one(ap, source, key, key_len, ctxt);
 	if (ret == NSS_STATUS_NOTFOUND) {
 		pthread_setcancelstate(cur_state, NULL);
 		return ret;
@@ -1012,11 +1003,11 @@ static int check_map_indirect(struct autofs_point *ap,
 	return NSS_STATUS_SUCCESS;
 }
 
-int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *context)
+int lookup_mount(struct autofs_point *ap, struct map_source *map, const char *name, int name_len, void *context)
 {
 	struct lookup_context *ctxt = (struct lookup_context *) context;
-	struct map_source *source;
-	struct mapent_cache *mc;
+	struct map_source *source = map;
+	struct mapent_cache *mc = source->mc;
 	struct mapent *me;
 	char key[KEY_MAX_LEN + 1];
 	int key_len;
@@ -1024,12 +1015,6 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	char mapent_buf[MAPENT_MAX_LEN + 1];
 	int ret;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
-	mc = source->mc;
-
 	debug(ap->logopt, MODPREFIX "looking up %s", name);
 
 	key_len = snprintf(key, KEY_MAX_LEN + 1, "%s", name);
@@ -1064,10 +1049,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		if (!lkp_key)
 			return NSS_STATUS_UNKNOWN;
 
-		master_source_current_wait(ap->entry);
-		ap->entry->current = source;
-
-		status = check_map_indirect(ap, lkp_key, strlen(lkp_key), ctxt);
+		status = check_map_indirect(ap, source, lkp_key, strlen(lkp_key), ctxt);
 		free(lkp_key);
 		if (status)
 			return status;
@@ -1114,11 +1096,9 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	if (!mapent)
 		return NSS_STATUS_TRYAGAIN;
 
-	master_source_current_wait(ap->entry);
-	ap->entry->current = source;
-
 	debug(ap->logopt, MODPREFIX "%s -> %s", key, mapent);
-	ret = ctxt->parse->parse_mount(ap, key, key_len,
+
+	ret = ctxt->parse->parse_mount(ap, source, key, key_len,
 				       mapent, ctxt->parse->context);
 	if (ret) {
 		/* Don't update negative cache when re-connecting */
