@@ -39,18 +39,18 @@ int parse_version = AUTOFS_PARSE_VERSION;	/* Required by protocol */
 
 static struct mount_mod *mount_nfs = NULL;
 static int init_ctr = 0;
-static pthread_mutex_t instance_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t parse_instance_mutex = PTHREAD_MUTEX_INITIALIZER;
 
-static void instance_mutex_lock(void)
+static void parse_instance_mutex_lock(void)
 {
-	int status = pthread_mutex_lock(&instance_mutex);
+	int status = pthread_mutex_lock(&parse_instance_mutex);
 	if (status)
 		fatal(status);
 }
 
-static void instance_mutex_unlock(void)
+static void parse_instance_mutex_unlock(void)
 {
-	int status = pthread_mutex_unlock(&instance_mutex);
+	int status = pthread_mutex_unlock(&parse_instance_mutex);
 	if (status)
 		fatal(status);
 }
@@ -112,7 +112,7 @@ int parse_init(int argc, const char *const *argv, void **context)
 
 	/* We only need this once.  NFS mounts are so common that we cache
 	   this module. */
-	instance_mutex_lock();
+	parse_instance_mutex_lock();
 	if (mount_nfs)
 		init_ctr++;
 	else {
@@ -121,11 +121,11 @@ int parse_init(int argc, const char *const *argv, void **context)
 		} else {
 			kill_context(ctxt);
 			*context = NULL;
-			instance_mutex_unlock();
+			parse_instance_mutex_unlock();
 			return 1;
 		}
 	}
-	instance_mutex_unlock();
+	parse_instance_mutex_unlock();
 
 	return 0;
 }
@@ -170,11 +170,9 @@ static struct substvar *add_lookup_vars(struct autofs_point *ap,
 
 	if (*key == '/')
 		strcpy(path, key);
-	else {
-		strcpy(path, ap->path);
-		strcat(path, "/");
-		strcat(path, key);
-	}
+	else
+		sprintf(path, "%s/%s", ap->path, key);
+
 	list = macro_addvar(list, "path", 4, path);
 
 	me = cache_lookup_distinct(source->mc, lkp_key);
@@ -425,6 +423,7 @@ static int sel_in_network(struct autofs_point *ap,
 	return ret;
 }
 
+#ifdef HAVE_INNETGR
 static int sel_netgrp(struct autofs_point *ap,
 		      struct selector *s, struct substvar *sv)
 {
@@ -489,6 +488,7 @@ out:
 
 	return ret;
 }
+#endif
 
 static int eval_selector(struct autofs_point *ap,
 			 struct amd_entry *this, struct substvar *sv)
@@ -628,7 +628,12 @@ static int eval_selector(struct autofs_point *ap,
 		switch (s->sel->selector) {
 		case SEL_NETGRP:
 		case SEL_NETGRPD:
+#ifndef HAVE_INNETGR
+			error(logopt, MODPREFIX
+			      "netgroups not available, function innetgr(3) not available");
+#else
 			ret = sel_netgrp(ap, s, sv);
+#endif
 			break;
 
 		default:
@@ -1067,24 +1072,23 @@ static int do_auto_mount(struct autofs_point *ap, const char *name,
 			 struct amd_entry *entry, unsigned int flags)
 {
 	char target[PATH_MAX + 1];
+	int len;
 
 	if (!entry->map_type) {
-		if (strlen(entry->fs) > PATH_MAX) {
+		len = snprintf(target, PATH_MAX, "%s", entry->fs);
+		if (len > PATH_MAX) {
 			error(ap->logopt, MODPREFIX
 			     "error: fs option length is too long");
 			return 0;
 		}
-		strcpy(target, entry->fs);
 	} else {
-		if (strlen(entry->fs) +
-		    strlen(entry->map_type) + 5 > PATH_MAX) {
+		len = snprintf(target, PATH_MAX,
+			       "%s,amd:%s", entry->map_type, entry->fs);
+		if (len > PATH_MAX) {
 			error(ap->logopt, MODPREFIX
 			     "error: fs + maptype options length is too long");
 			return 0;
 		}
-		strcpy(target, entry->map_type);
-		strcat(target, ",amd:");
-		strcat(target, entry->fs);
 	}
 
 	return do_mount(ap, ap->path,
@@ -1207,17 +1211,15 @@ static int do_nfs_mount(struct autofs_point *ap, const char *name,
 	char *opts = (entry->opts && *entry->opts) ? entry->opts : NULL;
 	unsigned int umount = 0;
 	int ret = 0;
+	int len;
 
-	if (strlen(entry->rhost) + strlen(entry->rfs) + 1 > PATH_MAX) {
+	len = snprintf(target, PATH_MAX, "%s:%s", entry->rhost, entry->rfs);
+	if (len > PATH_MAX) {
 		error(ap->logopt, MODPREFIX
 		     "error: rhost + rfs options length is too long");
 		return 1;
 	}
 
-	strcpy(target, entry->rhost);
-	strcat(target, ":");
-	strcat(target, entry->rfs);
-
 	proximity = get_network_proximity(entry->rhost);
 	if (proximity == PROXIMITY_OTHER && entry->remopts && *entry->remopts)
 		opts = entry->remopts;
@@ -1319,7 +1321,7 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 	 */
 	if (strcmp(name, entry->rhost)) {
 		char *target;
-		size_t len;
+		int len;
 
 		len = ap->len + strlen(entry->rhost) + 2;
 		target = malloc(len);
@@ -1328,9 +1330,7 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 			     "failed to alloc target to hosts mount base");
 			goto out;
 		}
-		strcpy(target, ap->path);
-		strcat(target, "/");
-		strcat(target, entry->rhost);
+		sprintf(target, "%s/%s", ap->path, entry->rhost);
 		if (entry->path)
 			free(entry->path);
 		entry->path = target;
@@ -1358,14 +1358,6 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 		argc = 1;
 	}
 
-	instance_mutex_lock();
-	status = open_lookup("hosts", MODPREFIX, NULL, argc, pargv, &lookup);
-	if (status != NSS_STATUS_SUCCESS) {
-		debug(ap->logopt, "open lookup module hosts failed");
-		instance_mutex_unlock();
-		goto out;
-	}
-
 	instance = master_find_source_instance(source,
 					 "hosts", "sun", argc, pargv);
 	if (!instance) {
@@ -1374,13 +1366,23 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 		if (!instance) {
 			error(ap->logopt, MODPREFIX
 			     "failed to create source instance for hosts map");
-			instance_mutex_unlock();
-			close_lookup(lookup);
 			goto out;
 		}
 	}
-	instance->lookup = lookup;
-	instance_mutex_unlock();
+
+	if (!instance->lookup) {
+		map_module_writelock(instance);
+		if (!instance->lookup) {
+			status = open_lookup("hosts", MODPREFIX, NULL, argc, pargv, &lookup);
+			if (status != NSS_STATUS_SUCCESS) {
+				map_module_unlock(instance);
+				debug(ap->logopt, "open lookup module hosts failed");
+				goto out;
+			}
+			instance->lookup = lookup;
+		}
+		map_module_unlock(instance);
+	}
 
 	cache_writelock(source->mc);
 	me = cache_lookup_distinct(source->mc, name);
@@ -1388,11 +1390,11 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 		cache_push_mapent(me, NULL);
 	cache_unlock(source->mc);
 
-	master_source_current_wait(ap->entry);
-	ap->entry->current = source;
-
-	ret = lookup->lookup_mount(ap, entry->rhost,
+	map_module_readlock(instance);
+	lookup = instance->lookup;
+	ret = lookup->lookup_mount(ap, instance, entry->rhost,
 				   strlen(entry->rhost), lookup->context);
+	map_module_unlock(instance);
 
 	if (!strcmp(name, entry->rhost))
 		goto out;
@@ -1814,7 +1816,7 @@ static void normalize_sublink(unsigned int logopt,
 			      struct amd_entry *entry, struct substvar *sv)
 {
 	char *new;
-	size_t len;
+	int len;
 
 	/* Normalizing sublink requires a non-blank fs option */
 	if (!*entry->fs)
@@ -1828,9 +1830,7 @@ static void normalize_sublink(unsigned int logopt,
 			      "error: couldn't allocate storage for sublink");
 			return;
 		}
-		strcpy(new, entry->fs);
-		strcat(new, "/");
-		strcat(new, entry->sublink);
+		sprintf(new, "%s/%s", entry->fs, entry->sublink);
 		debug(logopt, MODPREFIX
 		      "rfs dequote(\"%.*s\") -> %s",
 		      strlen(entry->sublink), entry->sublink, new);
@@ -1861,9 +1861,7 @@ static void update_prefix(struct autofs_point *ap,
 		len = strlen(ap->pref) + strlen(name) + 2;
 		new = malloc(len);
 		if (new) {
-			strcpy(new, ap->pref);
-			strcat(new, name);
-			strcat(new, "/");
+			sprintf(new, "%s%s/", ap->pref, name);
 			entry->pref = new;
 		}
 	}
@@ -2209,13 +2207,14 @@ done:
 	return defaults_entry;
 }
 
-int parse_mount(struct autofs_point *ap, const char *name,
-		int name_len, const char *mapent, void *context)
+int parse_mount(struct autofs_point *ap, struct map_source *map,
+		const char *name, int name_len, const char *mapent,
+		void *context)
 {
 	struct parse_context *ctxt = (struct parse_context *) context;
 	unsigned int flags = conf_amd_get_flags(ap->path);
 	struct substvar *sv = NULL;
-	struct map_source *source;
+	struct map_source *source = map;
 	unsigned int at_least_one;
 	struct list_head entries, *p, *head;
 	struct amd_entry *defaults_entry;
@@ -2225,10 +2224,6 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	int cur_state;
 	int ret;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
 	if (!mapent) {
 		warn(ap->logopt, MODPREFIX "error: empty map entry");
 		return 1;
@@ -2373,12 +2368,12 @@ int parse_done(void *context)
 	int rv = 0;
 	struct parse_context *ctxt = (struct parse_context *) context;
 
-	instance_mutex_lock();
+	parse_instance_mutex_lock();
 	if (--init_ctr == 0) {
 		rv = close_mount(mount_nfs);
 		mount_nfs = NULL;
 	}
-	instance_mutex_unlock();
+	parse_instance_mutex_unlock();
 	if (ctxt)
 		kill_context(ctxt);
 
