@@ -41,18 +41,18 @@ int parse_version = AUTOFS_PARSE_VERSION;	/* Required by protocol */
 
 static struct mount_mod *mount_nfs = NULL;
 static int init_ctr = 0;
-static pthread_mutex_t instance_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t parse_instance_mutex = PTHREAD_MUTEX_INITIALIZER;
 
-static void instance_mutex_lock(void)
+static void parse_instance_mutex_lock(void)
 {
-	int status = pthread_mutex_lock(&instance_mutex);
+	int status = pthread_mutex_lock(&parse_instance_mutex);
 	if (status)
 		fatal(status);
 }
 
-static void instance_mutex_unlock(void)
+static void parse_instance_mutex_unlock(void)
 {
-	int status = pthread_mutex_unlock(&instance_mutex);
+	int status = pthread_mutex_unlock(&parse_instance_mutex);
 	if (status)
 		fatal(status);
 }
@@ -82,7 +82,6 @@ static struct parse_context default_context = {
 	1			/* Do slashify_colons */
 };
 
-int destroy_logpri_fifo(struct autofs_point *ap);
 static char *concat_options(char *left, char *right);
 
 /* Free all storage associated with this context */
@@ -376,10 +375,16 @@ static int do_init(int argc, const char *const *argv, struct parse_context *ctxt
 	if (gbl_options) {
 		append_options = defaults_get_append_options();
 		if (append_options) {
-			char *tmp = concat_options(gbl_options, ctxt->optstr);
+			char *tmp;
+
+			errno = 0;
+			tmp = concat_options(gbl_options, ctxt->optstr);
 			if (!tmp) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				logerr(MODPREFIX "concat_options: %s", estr);
+				/* Ignore non-error NULL return */
+				if (errno) {
+					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+					logerr(MODPREFIX "concat_options: %s", estr);
+				}
 				/* freed in concat_options */
 				ctxt->optstr = NULL;
 			} else
@@ -424,7 +429,7 @@ int parse_init(int argc, const char *const *argv, void **context)
 
 	/* We only need this once.  NFS mounts are so common that we cache
 	   this module. */
-	instance_mutex_lock();
+	parse_instance_mutex_lock();
 	if (mount_nfs)
 		init_ctr++;
 	else {
@@ -432,11 +437,11 @@ int parse_init(int argc, const char *const *argv, void **context)
 			init_ctr++;
 		} else {
 			kill_context(ctxt);
-			instance_mutex_unlock();
+			parse_instance_mutex_unlock();
 			return 1;
 		}
 	}
-	instance_mutex_unlock();
+	parse_instance_mutex_unlock();
 
 	*context = (void *) ctxt;
 
@@ -782,7 +787,14 @@ static int check_is_multi(const char *mapent)
 		if (not_first_chunk) {
 			if (*p == '"')
 				p++;
-			if (*p == '/' || *p == '-') {
+			/*
+			 * Although an options string here would mean
+			 * we have a multi-mount we can't rely on it
+			 * since it's also valid in a mount location.
+			 */
+			if (*p == '-')
+				p++;
+			if (*p == '/') {
 				multi = 1;
 				break;
 			}
@@ -929,6 +941,12 @@ static int validate_location(unsigned int logopt, char *loc)
 	if (*ptr == ':')
 		return 1;
 
+	/* Fail on replicated entry with empty first host name */
+	if (*ptr == ',') {
+		error(logopt, "missing first host name in location %s", loc);
+		return 0;
+	}
+
 	/*
 	 * If a ':/' is present now it must be a host name, except
 	 * for those special file systems like sshfs which use "#"
@@ -965,6 +983,18 @@ static int validate_location(unsigned int logopt, char *loc)
 				      "found in location %s", *ptr, loc);
 				return 0;
 			}
+
+			/* Fail on replicated entry with empty host name */
+			if (*ptr == ',') {
+				char next = *(ptr + 1);
+
+				if (next == ',' || next == ':') {
+					error(logopt,
+					      "missing host name in location %s", loc);
+					return 0;
+				}
+			}
+
 			ptr++;
 		}
 
@@ -1007,9 +1037,12 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 				free(myoptions);
 				myoptions = newopt;
 			} else if (newopt) {
+				errno = 0;
 				tmp = concat_options(myoptions, newopt);
-				if (!tmp) {
+				/* Ignore non-error NULL return */
+				if (!tmp && errno) {
 					char *estr;
+
 					estr = strerror_r(errno, buf, MAX_ERR_BUF);
 					error(logopt, MODPREFIX
 					      "concat_options: %s", estr);
@@ -1142,6 +1175,9 @@ static int mount_subtree(struct autofs_point *ap, struct mapent_cache *mc,
 		if (!len) {
 			warn(ap->logopt, "path loo long");
 			cache_unlock(mc);
+			cache_writelock(mc);
+			tree_mapent_delete_offsets(mc, name);
+			cache_unlock(mc);
 			return 1;
 		}
 		key[len] = '/';
@@ -1186,6 +1222,9 @@ static int mount_subtree(struct autofs_point *ap, struct mapent_cache *mc,
 				cache_unlock(mc);
 				error(ap->logopt, MODPREFIX
 					 "failed to mount offset triggers");
+				cache_writelock(mc);
+				tree_mapent_delete_offsets(mc, name);
+				cache_unlock(mc);
 				return 1;
 			}
 		}
@@ -1278,13 +1317,14 @@ static void cleanup_offset_entries(struct autofs_point *ap,
  * level nexting point. Finally to mount non multi-mounts and to mount a
  * lower level multi-mount nesting point and its offsets.
  */
-int parse_mount(struct autofs_point *ap, const char *name,
-		int name_len, const char *mapent, void *context)
+int parse_mount(struct autofs_point *ap, struct map_source *map,
+		const char *name, int name_len, const char *mapent,
+		void *context)
 {
 	struct parse_context *ctxt = (struct parse_context *) context;
 	char buf[MAX_ERR_BUF];
-	struct map_source *source;
-	struct mapent_cache *mc;
+	struct map_source *source = map;
+	struct mapent_cache *mc = source->mc;
 	struct mapent *me, *oe, *tmp;
 	LIST_HEAD(offsets);
 	char *pmapent, *options;
@@ -1294,12 +1334,6 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	int slashify = ctxt->slashify_colons;
 	unsigned int append_options;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
-	mc = source->mc;
-
 	if (!mapent) {
 		warn(ap->logopt, MODPREFIX "error: empty map entry");
 		return 1;
@@ -1375,8 +1409,10 @@ dont_expand:
 				free(mnt_options);
 				mnt_options = noptions;
 			} else if (noptions) {
+				errno = 0;
 				tmp = concat_options(mnt_options, noptions);
-				if (!tmp) {
+				/* Ignore non-error NULL return */
+				if (!tmp && errno) {
 					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 					error(ap->logopt,
 					      MODPREFIX "concat_options: %s", estr);
@@ -1400,8 +1436,10 @@ dont_expand:
 				free(options);
 				options = mnt_options;
 			} else if (mnt_options) {
+				errno = 0;
 				tmp = concat_options(options, mnt_options);
-				if (!tmp) {
+				/* Ignore non-error NULL return */
+				if (!tmp && errno) {
 					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 					error(ap->logopt, MODPREFIX "concat_options: %s", estr);
 					free(pmapent);
@@ -1722,12 +1760,12 @@ int parse_done(void *context)
 	int rv = 0;
 	struct parse_context *ctxt = (struct parse_context *) context;
 
-	instance_mutex_lock();
+	parse_instance_mutex_lock();
 	if (--init_ctr == 0) {
 		rv = close_mount(mount_nfs);
 		mount_nfs = NULL;
 	}
-	instance_mutex_unlock();
+	parse_instance_mutex_unlock();
 	if (ctxt)
 		kill_context(ctxt);
 
